__all__ = [
    'send_sns'
]

from ._imports import *
from ._config import config
from ._awsbase import Aws_base

class Amazonsns(Aws_base):
    clients = {}

    @classmethod
    def get_client(cls, env):
        return super()._get_client(env, 'sns')
    
def snsclient(env):
    return Amazonsns.get_client(env)

@atexit.register
def sns_close():
    for client in Amazonsns.clients.values():
        client.close()


def send_sns(sub, msg, env):
    '''
    env = 'dev'
    sns_sub =  f"Data science team cash flow job status in {env}"
    sns_msg = tabulate(df_job_status, headers='keys', tablefmt='psql') #use tablulate to give table str format
    send_sns(sns_sub, sns_msg, env)  
    '''    
    if env == 'dev':
        arn = r'arn:aws:sns:us-east-2:392523246457:amica_eds_ds_dev' 
    elif env == 'prod':
        arn = r'arn:aws:sns:us-east-2:453362238738:amica_eds_ds_prod'
    response = snsclient(env).publish(
    TopicArn=arn,
    Message= msg,
    Subject= sub
    )




__all__ = ['aws_athena_query']

from ._imports import *
from ._config import *
from ._amazons3 import s3_convert_file_to_df

class Aws_athena:

    def __init__(self, db, query, env) -> None:
        self.param = self._get_athena_params(db, query, env)
    
    def _get_athena_params(self, db, query, env):
        param = {}
        param ['query'] = query
        param ['athena db'] =  db  
        param['env'] = env
        param ['session'] =  boto3.Session()       
        param ['athena output folder'] = f's3://amica-aws-eds-{env}-datascience/athena_temp/' if env !='dl' else f's3://amica-eds-edsc/athena_temp/'
        param ['athena output bucket'] = param ['athena output folder'].split('/')[2]
        param ['athena output prefix'] = 'athena_temp'
        #param ['athena output bucket S3 path'] = r"s3://" + param ['athena output bucket'] + '/'        
        return param 
    
    def get_athena_query(self):        
        env = self.param['env']
        try:        
            athena_client = self.param ['session'].client('athena', region_name='us-east-2',
                                        aws_access_key_id=config[f'aws_{env}']['aws_access_key_id'],
                                        aws_secret_access_key=config[f'aws_{env}']['aws_secret_access_key'])
            
            response = athena_client.start_query_execution(
                    QueryString = self.param['query'],
                        QueryExecutionContext={
                        'Database': self.param['athena db']
                        },
                        ResultConfiguration={
                        'OutputLocation': self.param['athena output folder'],
                        }
                )            
            self.filename = response['QueryExecutionId'] 
            query_status = None
            while query_status == 'QUEUED' or query_status == 'RUNNING' or query_status is None:
                query_status = athena_client.get_query_execution(QueryExecutionId=self.filename)['QueryExecution']['Status']['State']
                if query_status == 'FAILED' or query_status == 'CANCELLED':
                    raise Exception('Athena query failed or was cancelled')
                time.sleep(3)
            #query complete  
            key= self.param ['athena output prefix'] + '/' + self.filename + '.csv'                  
            df = s3_convert_file_to_df(self.param ['athena output bucket'], key , env)            
            return df
        except Exception as e:
            print(e) 

    
    
    def athena_cleanup(self, env):
        s3 = self.param ['session'].resource('s3',                               
            aws_access_key_id=config[f'aws_{env}']['aws_access_key_id'],
            aws_secret_access_key=config[f'aws_{env}']['aws_secret_access_key'],           
                        )
        my_bucket = s3.Bucket(self.param['athena output bucket'])
        for item in my_bucket.objects.filter(Prefix=self.param['athena output prefix']):
            item.delete()        
        #restore the athena_temp folder
        object = s3.Object(self.param['athena output bucket'], self.param['athena output prefix'])
        object.put()

def aws_athena_query(db, query, env):
    '''
    get data in df from athena query
    exp1
    db = 'rravichandran_test_db'
    query = 'select * FROM "rravichandran_test_db"."documents"'
    env = 'dev'
    df = aws_athena_query(db, query, env)
    exp2
    db = 'eds_ds_ptc'
    query =  'SELECT * FROM "eds_ds_ptc"."prediction_history" limit 10;'
    env = 'dl'
    aws_athena_query(db, query, env)
    '''
    aa = Aws_athena(db, query, env)
    df = aa.get_athena_query()
    aa.athena_cleanup(env)  #clean temp folder    An error occurred (AccessDenied) when calling the DeleteObject operation: Access Denied
    return df


__all__ = ['admission', 'train_pipeline']

from ._imports import *
from ._pconfig import *
from ._features import *
from ._misc import *

def admission():
    print('testing...')

def train_pipeline(train_start, train_end, env='dev', logger=None, automl=True, update_model=False):
    error_status = 0
    if logger:
        logger.info(f'Train from {train_start} to {train_end}.')
    try:
        runtime_train = create_train_runtime(train_start, train_end, env=env, logger=logger)
        run_months = get_run_months(train_start, train_end)

        df = pd.DataFrame()
        for run_month in run_months:
            runtime = create_runtime_per_runmonth(run_month, env, logger)
            df_base = dataset_initialize(runtime)
            df_feat = dataset_features(df_base, runtime)
            df_targ = dataset_target(df_feat, runtime)
            df_filt = dataset_filter(df_targ, runtime)
            df = pd.concat([df, df_filt], ignore_index=True)
            if logger:
                logger.info(f'Run date {run_month} with {len(df_targ)} records.')

        if automl:
            evals = dataset_retrain(df, runtime_train, automl=automl_flaml, automl_kwargs={'time_budget': 60*60})
        else:
            evals = dataset_retrain(df, runtime_train)
        save_output_mstr(evals, runtime_train)

        if update_model:
            update_model_version(env, promote=True)
    except Exception as e:
        if logger:
            logger.error('Error', exc_info=e)
        error_status = 1
    finally:
        return error_status

def create_runtime_per_runmonth(run_month, env, logger, run_day=pconfig['run_day']):
    features = Features()
    # features.remove_features([c for c in features.get_feature_dict() if c.startswith('PC') or c.startswith('CC')]) # debugging for now
    features.remove_features(['CC_CLAIM_NOTES'])
    #remove demo
    demo_feature = ['GOOD_STUDENT','NUM_CONVICTIONS','DRIVER_AGE_DISC','DRIVER_TRAINING','YOUTHFUL_OPER','STUDENT_AWAY','MILITARY_IND','RIDER_TRAINING','HOUSEHOLD_INCOME']
    features.remove_features(demo_feature)

    runtime = {}
    runtime['timestamp']      = dt.datetime.now().strftime("%Y%m%d%H%M%S")
    runtime['env']            = env
    runtime['logger']         = logger

    runtime['features']       = features.get_feature_dict()
    runtime['teradata_table'] = f'{config["teradata"]["user"]}.temp_table'
    runtime['type_dt']        = pconfig['type_dt']

    run_month = dt.datetime.strptime(run_month, '%Y-%m')
    runtime['run_month']       = run_month
    runtime['run_month_start'] = DtFind.first_day_of_month(run_month).strftime('%Y-%m-%d')
    runtime['run_month_end']   = DtFind.last_day_of_month(run_month).strftime('%Y-%m-%d')
    runtime['run_dt']          = DtFind.first_day_last_month(run_month).replace(day=run_day).strftime('%Y-%m-%d')

    return runtime

def create_train_runtime(run_start, run_end, env, logger):
    runtime                   = {}
    runtime['timestamp']      = dt.datetime.now().strftime("%Y%m%d%H%M%S")
    runtime['env']            = env
    runtime['logger']         = logger
    runtime['teradata_table'] = f'{config["teradata"]["user"]}.temp_table'
    runtime['run_start']      = run_start
    runtime['run_end']        = run_end
    return runtime 

def get_run_months(run_start, run_end, dt_format='%Y-%m'):
    a = dt.datetime.strptime(run_start, dt_format)
    b = dt.datetime.strptime(run_end, dt_format)
    run_months = [a.strftime(dt_format)]
    while a.strftime(dt_format) != b.strftime(dt_format):
        a = DtFind.first_day_next_month(a)
        run_months.append(a.strftime(dt_format))
    return run_months

def pc_initialize(df):
    pc_temp(df['POLICY_NO'].rename('policy_number'), '#tmp_policynumber')
    pc_temp(df['ACCOUNT_NO'].rename('Amica_number'), '#tmp_amicanumber')

def dataset_initialize(runtime):
    table_name  = runtime['teradata_table']
    table_query = f'''
        select 
            trim(a.policy_key) as POLICY_KEY
            , trim(a.policy_no) as POLICY_NO
            , trim(a.account_no) as ACCOUNT_NO
            , trim(a.perpetual_pol_no) as PERPETUAL_POL_NO
            , trim(a.expired_policy_no) as EXPIRED_POLICY_NO
            , a.effective_date as EFFECTIVE_DATE
            , a.expiration_date as EXPIRATION_DATE
            , a.process_ts as PROCESS_TS
        from atuprv.dw_agreement a
        left join atuprv.dw_state b on a.iso_state_number = b.iso_state_number
        left join atuprv.dw_form_desc c on a.form_cd = c.code
        where 
            a.{runtime["type_dt"]}_date between date \'{runtime["run_month_start"]}\' and date \'{runtime["run_month_end"]}\'
            and cast(a.process_ts as Date) <= \'{runtime["run_dt"]}\'
            and a.policy_type = 1 -- auto line of business
            and a.policy_activity = 'Renewal'
        qualify row_number() over (partition by a.policy_key order by a.process_ts desc)=1
    '''

    # TERADATA TEMP TABLE
    td_drop_table(table_name)
    td_volatile(table_query, table_name)
    df = td_query(query_creator(table_columns='*', table_name=table_name, table_ops=''))

    # POLICY CENTER TEMP TABLE
    #pc_initialize(df)
    return df

def dataset_filter(df, runtime):
    if runtime['features']['SOURCE_OF_BUSINESS']:
        df = df.loc[df['SOURCE_OF_BUSINESS']!='MAIP POLICY']
    if runtime['features']['STATE']:
        df = df.loc[
            (df['STATE_CODE'].isin(
                ('AK', 'AL', 'AR', 'AZ', 'CA', 'CO', 'CT', 'DC', 'DE', 'FL', 'GA', 'HI', 'IA', 'ID', 'IL', 'IN', 'KS', 'KY', 'LA', 'MA', 
                'MD', 'ME', 'MI', 'MN', 'MO', 'MS', 'MT', 'NC', 'ND', 'NE', 'NH', 'NJ', 'NM', 'NV', 'NY', 'OH', 'OK', 'OR', 'PA', 'RI', 'SC', 'SD', 'TN', 
                'TX', 'UT', 'VA', 'VT', 'WA', 'WI', 'WV', 'WY')
            ))
        ]
    if runtime['features']['POLICY_FORM']:
        df = df.loc[
            (df['POLICY_FORM'].isin(
                ('MOTORCYCLE', 'TRAILER', 'SNOWMOBILE', 'MANUAL AUTO', 'ANTIQUE AUTO', 'PRIVATE PASSENGER', 'MOBILE HOME', 'MOTOR HOME', 'GOLF CART')
            ))
        ]
    return df

def dataset_target(df, runtime):
    if not runtime['features']['CANCEL_RENEWAL'] or not runtime['features']['CANCEL_EXPIRING']:
        return df

    def target_cancel(row):
        cancel_reasons = [
            'RatesAnotherCarrier', 
            'nonpayment', 
            'EffortsDeterReasonFailed', 
            'RatesAnotherCarrierOutJur', 
            'PlacedLocalAgentSpouseCo',
        ]
        target = (
            (
                row['CANCEL_POST_RENEWAL_DAYS'] < np.inf
            ) and (
                (
                    row['CANCEL_RENEWAL_REASON'] in cancel_reasons 
                    and 
                    row['CANCEL_POST_RENEWAL_DAYS'] <= pconfig['target_post_window']
                ) or (
                    row['CANCEL_EXPIRING_REASON'] in cancel_reasons 
                    and 
                    row['CANCEL_PRE_RENEWAL_DAYS'] <= pconfig['target_pre_window']
                )
            )
        )
        return target

    df[pconfig['target']] = (
        df
        .assign(
            CANCEL_PRE_RENEWAL_DAYS = (
                pd.to_datetime(df['EFFECTIVE_DATE'], errors='coerce')
                - 
                pd.to_datetime(df['CANCEL_EXPIRING_DATE'], errors='coerce')
            ).dt.days.fillna(np.inf), 
            CANCEL_POST_RENEWAL_DAYS = (
                pd.to_datetime(df['CANCEL_RENEWAL_DATE'], errors='coerce')
                - 
                pd.to_datetime(df['EFFECTIVE_DATE'], errors='coerce')
            ).dt.days.fillna(np.inf),
        )
        .apply(target_cancel, axis=1).astype(int)
    )
    return df

def dataset_retrain(df, runtime, holdout_size=0.2, valid_size=0.2, random_state=42, resample=False, automl=None, automl_kwargs={}):
    model_config = load_model_config(runtime)
    model        = load_model(model_config, runtime)

    train, holdout = sklearn.model_selection.train_test_split(
        df, test_size=holdout_size, random_state=random_state, stratify=df[pconfig['target']]
    )
    train, valid   = sklearn.model_selection.train_test_split(
        train, test_size=valid_size, random_state=random_state,
    )

    train   = train[model_config['features'] + [pconfig['target']]]
    holdout = holdout[model_config['features'] + [pconfig['target']]]
    valid   = valid[model_config['features'] + [pconfig['target']]]

    if automl:
        estimator = automl(
            train[model_config['features']], train[pconfig['target']],
            valid[model_config['features']], valid[pconfig['target']],
            **automl_kwargs
        )
        model_retrain = sklearn.pipeline.make_pipeline(model[-2], estimator)
    else:
        model_retrain = sklearn.base.clone(model)
    

    if resample:
        train = undersample_target(train)

    model_retrain.fit(train[model_config['features']], train[pconfig['target']])
    valid_probs   = model_retrain.predict_proba(valid[model_config['features']])[:,1]
    holdout_probs = model_retrain.predict_proba(holdout[model_config['features']])[:,1]

    thr           = get_threshold(valid[pconfig['target']], valid_probs, type='f1')
    holdout_preds = (holdout_probs>thr).astype(int)
    evals         = evaluate_metrics(holdout[pconfig['target']], holdout_probs, holdout_preds)
    importances   = feature_importance(train, model_retrain, model_config)

    model_version_retrain = f'{model_config["model_version"]}.retrain'
    model_version_update  = str(int(model_config["model_version"]) + 1)
    model_config_retrain  = {
        'model_name':   'propensity_to_cancel',
        'model_version': model_version_update,
        'model_type':    model_retrain[-1].__class__.__name__,
        'model_file':    f'cancel_v{model_version_retrain}.pkl',
        'features':      model_config['features'],
        'threshold':     str(np.round(thr, 3)),
        'evaluation':    evals,
        'importances':   importances,
    }
    model_ref_retrain = holdout.assign(
        PROBABILITY = holdout_probs,
        PREDICTION  = holdout_preds,
    )

    #save_model_assets('dev', model_retrain, model_config_retrain, model_ref_retrain, model_version_retrain, model_data=df)
    save_model_assets(runtime['env'], model_retrain, model_config_retrain, model_ref_retrain, model_version_retrain, model_data=df)
    
    return evals

def undersample_target(df):
    df_balanced = df.groupby(df[pconfig['target']])
    df_balanced = df_balanced.apply(lambda x: 
        x.sample(df_balanced.size().min())
    ).reset_index(drop=True)
    return df_balanced

def preprocess_features(model_config, train, holdout=None, valid=None):
    txt_columns = ['CLAIM_NOTES']
    cat_columns = train.select_dtypes(include=['object', 'category']).columns.to_list()
    num_columns = train.select_dtypes(include='number').columns.to_list()
    
    cat_columns = [c for c in cat_columns if c in model_config['features'] and c not in txt_columns]
    num_columns = [c for c in num_columns if c in model_config['features'] and c not in txt_columns]

    pipe_num = sklearn.pipeline.make_pipeline(
        sklearn.impute.SimpleImputer(strategy='median'), 
        sklearn.preprocessing.MinMaxScaler(),
     )
    pipe_cat = sklearn.pipeline.make_pipeline(
        sklearn.impute.SimpleImputer(strategy='constant', fill_value='Missing'), 
        sklearn.preprocessing.OneHotEncoder(handle_unknown='ignore'),
    )

    transformer = sklearn.compose.make_column_transformer(
        (pipe_num, num_columns),
        (pipe_cat, cat_columns),
        remainder='passthrough',
    )

    X_train = transformer.fit_transform(train[model_config['features']])
    X_train = pd.DataFrame(X_train, columns=transformer.get_feature_names_out())
    if valid is not None: 
        X_valid = transformer.transform(valid[model_config['features']])
        X_valid = pd.DataFrame(X_valid, columns=transformer.get_feature_names_out())
    if holdout is not None: 
        X_holdout = transformer.transform(holdout[model_config['features']])
        X_holdout = pd.DataFrame(X_holdout, columns=transformer.get_feature_names_out())
    return X_train, X_valid, X_holdout

def evaluate_metrics(actual, prob, pred):
    evals = {
        'support':     len(actual),
        'accuracy':    sklearn.metrics.accuracy_score(actual, pred).round(2),
        'precision':   sklearn.metrics.precision_score(actual, pred).round(2),
        'recall':      sklearn.metrics.recall_score(actual, pred).round(2),
        'specificity': sklearn.metrics.recall_score(actual, pred, pos_label=0).round(2),
        'f1':          sklearn.metrics.f1_score(actual, pred).round(2),
        'rocauc' :     sklearn.metrics.roc_auc_score(actual, prob).round(2),
    }
    return evals

def get_threshold(test_y, test_probs, type='roc_auc'):
    if type == 'roc_auc':
        fpr, tpr, thr = metrics.roc_curve(test_y, test_probs)
        thr = thr[np.argmax(tpr-fpr)]
    else: # type == 'f1'
        precision, recall, thr = metrics.precision_recall_curve(test_y, test_probs)
        precision_filter = precision[(precision>0) & (recall>0)]
        recall_filter = recall[(precision>0) & (recall>0)]
        precision = precision_filter
        recall = recall_filter
        thr = thr[np.argmax(precision*recall/(precision+recall))]
    return thr

def feature_importance(df, model, model_config, n=10000, random_state=100):
    permute_importance = {}
    n_iter = 10

    feats  = model_config['features']
    thr    = float(model_config['threshold'])

    sample = df.sample(n=min(n, len(df)), random_state=random_state)
    sample_probs = model.predict_proba(sample[feats])[:,1]
    sample_preds = (sample_probs>thr).astype(int)
    sample_acc   = metrics.accuracy_score(sample[pconfig['target']], sample_preds)
    for f in feats:
        permute_acc_n = []
        for n in range(n_iter):
            permute_feats = sample[feats].copy(deep=True)
            permute_feats[f] = permute_feats[f].sample(frac=1, random_state=n).values
            permute_probs = model.predict_proba(permute_feats)[:,1]
            permute_preds = (permute_probs>thr).astype(int)
            permute_acc = metrics.accuracy_score(sample[pconfig['target']], permute_preds)
            permute_acc_n.append(permute_acc)
        permute_importance[f] = sample_acc - np.mean(permute_acc_n)
    permute_importance = pd.Series(permute_importance)
    permute_importance = permute_importance / permute_importance.max()
    permute_importance = permute_importance.clip(lower=0).sort_values(ascending=False)
    return permute_importance.to_dict()

def load_model_config(runtime):
    model_version           = get_model_version(runtime["env"])
    local_model_config_file = os.path.join(config['local']['local_folder'], 'model_config.json')
    s3_model_config_file    = os.path.join(pconfig[f's3_model_assets_folder_{runtime["env"]}'], f'cancel_v{model_version}.json')
    s3_download_file(s3_model_config_file, local_model_config_file, env=runtime['env'])

    with open(local_model_config_file, 'r') as f:
        model_config = json.load(f)
    os.remove(local_model_config_file)
    return model_config

def load_model(model_config, runtime):
    local_model_pickle_file = os.path.join(config['local']['local_folder'], 'model.pkl')
    s3_model_pickle_file    = os.path.join(pconfig[f's3_model_assets_folder_{runtime["env"]}'], model_config['model_file'])
    s3_download_file(s3_model_pickle_file, local_model_pickle_file, env=runtime['env'])

    with open(local_model_pickle_file, 'rb') as f:
        model = pickle.load(f)
    os.remove(local_model_pickle_file)
    return model
    
def save_model_assets(env, model, model_config, model_ref, model_version, model_data=None, overwrite=True):
    s3_save_and_upload_file(
        model_config,
        s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{env}'], f'cancel_v{model_version}.json'),
        overwrite=overwrite,
        env=env,
    )
    s3_save_and_upload_file(
        model,
        s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{env}'], f'cancel_v{model_version}.pkl'),
        overwrite=overwrite,
        env=env,
    )
    s3_save_and_upload_file(
        model_ref,
        s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{env}'], f'cancel_v{model_version}.csv'),
        overwrite=overwrite,
        env=env,
    )
    if model_data is not None:
        s3_save_and_upload_file(
            model_data,
            s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{env}'], f'data_cancel_v{model_version}.csv'),
            overwrite=overwrite,
            env=env,
        )

def save_output_mstr(data, runtime):
    s3_save_and_upload_json(
        data,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
            'retrain/',
            f'cancel_train_{runtime["run_start"]}_{runtime["run_end"]}_{runtime["timestamp"]}.json'
        ),
        env=runtime['env'],
    )

def automl_flaml(train_x, train_y, val_x, val_y, time_budget=60):
    weight  = train_y.value_counts()[0]/train_y.value_counts()[1]
    
    custom_hp = {
        "xgb_limitdepth": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "scale_pos_weight":{
                "domain": weight,
                "init_value": weight,
            },
        },
        "xgboost": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "scale_pos_weight":{
                "domain": weight,
                "init_value": weight,
            },
        },
        "rf": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "class_weight": {
                "domain":'balanced',
                "init_value":'balanced',
            },
        },
        "extra_tree": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "class_weight": {
                "domain":'balanced',
                "init_value":'balanced',
            },
        },
        "catboost": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "scale_pos_weight": {
                "domain": weight,
                "init_value": weight,
            },
        },
        "lgbm": {
            "n_estimators": {
                "domain": flaml.tune.lograndint(lower=10, upper=200),
                "low_cost_init_value": 10,
            },
            "num_leaves": {
                "domain": flaml.tune.lograndint(lower=5, upper=100),
                "low_cost_init_value": 5,
            },
            "scale_pos_weight": {
                "domain": weight,
                "init_value": weight,
            },
        },
    }
    clf_settings = {
        'time_budget': time_budget,
        'metric': 'log_loss',
        'task': 'classification',
        'estimator_list': [
                'rf', 
                'xgboost',
                'xgb_limitdepth', 
                'extra_tree',  
                'lgbm',
            ],
        'verbose': 0,
        'custom_hp': custom_hp,
    }
    clf_dict = {
            'lgbm':             lightgbm.LGBMClassifier, 
            'rf':               sklearn.ensemble.RandomForestClassifier, 
            'xgboost':          xgboost.XGBClassifier, 
            'xgb_limitdepth':   xgboost.XGBClassifier, 
            'extra_tree':       sklearn.ensemble.ExtraTreesClassifier, 
        }

    automl = flaml.AutoML()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        automl.fit(
                X_train=train_x, y_train=train_y, 
                X_val=val_x, y_val=val_y,
                **clf_settings
            )
    clf = clf_dict[automl.best_estimator]
    clf_params = automl.best_config
    if automl.best_estimator not in ['xgboost', 'xgb_limitdepth']:
        clf_params = {p:clf_params[p] for p in clf_params if p in clf.__init__.__code__.co_varnames}
    return clf(**clf_params)




from ._get_geo_dataset import *
from ._wind import *
from ._imports import *
from ._pconfig import pconfig

data_folder, output_folder = pconfig['data_folder'], pconfig['output_folder']

def get_claim_data(cat_no, cat_date):

    sql_claim = f'''
        declare @lossdate date = '{cat_date }'

    drop table if exists #tmp_claim     
    ;with cte as (
    select 
    CAT_NO, 
    POLICY_NO,
    [FILE_NO],
    [FILETYPE],
    LOSS_DT,
    POL_1ST_INS_DATE,
    RES_PREM_ADDRESS,
    CITY,
    STATE, 
    ZIP5,
    COUNTY, 
    EXP_AMT,   --what is this 
    COVA,          --  what is this value? 21718.15 in thsi view but show      limit 706000  select  COV_A_LIMIT from POLICY_HISTORY_VIEW where policy_no = '66024221VG'
    COVA_DED,  --what is this
    COVB,
    COVB_DED,  --what is this
    --COVC,
    --COVC_DED,
    --COVD,
    --COVD_DED,
    GEO_SCORE,
    GEO_MATCH_ADDR,
    GEO_X,
    GEO_Y,
    GEO_ADDR_TYPE,
    GEO_SPATIAL_GEOM,  --is this geo code?
    FILE_HDLR_BRANCH,
    row_number() over (partition by policy_no order by cova desc ) as rnk  --some record policy dup  '6512421122' 
    from 
    CLAIMS_HO_HISTPOLICY 
    where LOB = 'Home'
    and COV = 'HAIL'
    and CAT_NO  = '{cat_no}'
    and loss_dt = @lossdate
    )

    select
    CAT_NO, 
    POLICY_NO,
    [FILE_NO],
    [FILETYPE],
    LOSS_DT,
    POL_1ST_INS_DATE,
    RES_PREM_ADDRESS,
    CITY,
    STATE, 
    ZIP5,
    COUNTY, 
    EXP_AMT,   --what is this 
    COVA,          --  what is this value? 21718.15 in thsi view but show      limit 706000  select  COV_A_LIMIT from POLICY_HISTORY_VIEW where policy_no = '66024221VG'
    COVA_DED,  --what is this
    COVB,
    COVB_DED,  --what is this
    GEO_SCORE,
    GEO_MATCH_ADDR,
    GEO_X,
    GEO_Y,
    GEO_ADDR_TYPE,
    GEO_SPATIAL_GEOM,  --is this geo code?
    FILE_HDLR_BRANCH
    into #tmp_claim
    from cte
    where rnk =1 


    --select * from #tmp_claim

    --get account number
    drop table if exists #tmp_claim_with_account

    select 
    b.ACCOUNT_NO,
    b.POL_1ST_INS_DATE as POL_1ST_INS_DATE_pol, 
    b.GEO_X as geo_x_pol,
    b.GEO_Y as geo_y_pol,
    a.*
    into #tmp_claim_with_account
    from  #tmp_claim a      
    join POLICY_HISTORY_VIEW b  on a.POLICY_NO = b.POLICY_NO  and a.GEO_X =b.GEO_X and a.GEO_Y = b.GEO_Y     
    where 
    b.INFORCE_BEGIN_DT <= @lossdate AND b.INFORCE_END_DT >@lossdate

    select * from #tmp_claim_with_account
    '''
    df_claim = gis_query(sql_claim)
    return df_claim


def get_all_policies_span_over_date(cat_date):
    policy_query= f'''
        select 
        ACCOUNT_NO, 
        POLICY_NO,
        POLICY_EFF_DATE,
        INFORCE_BEGIN_DT,
        INFORCE_END_DT,
        RES_PREM_ADDRESS,
        [CITY],
        [STATE],
        [ZIP5],
        COV_A_LIMIT,
        COV_A_LIMIT,
        LINE_OF_BUSINESS,
        [GEO_X],
        [GEO_Y]
        from 
        POLICY_HISTORY_VIEW 
        Where INFORCE_BEGIN_DT <= '{cat_date}' AND INFORCE_END_DT >'{cat_date}'     --the latest is 2023-10-31 not updated after that
        and LINE_OF_BUSINESS = 'HOMEOWNERS'          
        order by INFORCE_BEGIN_DT desc
'''
    df_policy = gis_query(policy_query)
    return df_policy


# def get_policy_outside_swath (gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, df_policy, gdf_has_no_claim_in_hailswath):
#     cat_zipcode  = set(list(gdf_has_claim_inside_swath['ZIP5'].unique()) + list(gdf_has_claim_not_in_swath['ZIP5'].unique()) )
#     df_policy_outside_swath = df_policy[df_policy['ZIP5'].isin(cat_zipcode) ]  #this minimize the policy number to be matched with claim data, and also get the policy data outside swath but in the same zipcode, which can be used as control group for later modeling.
#     df_has_no_claim_policy_not_in_swath = df_policy_outside_swath[~df_policy_outside_swath['POLICY_NO'].isin(gdf_has_no_claim_in_hailswath['POLICY_NO'])]
#     return df_policy_outside_swath, df_has_no_claim_policy_not_in_swath


# def get_outside_swath_nearest_polygon(df_has_no_claim_policy_not_in_swath, gdf):
#     geometry = [Point(xy) for xy in zip(df_has_no_claim_policy_not_in_swath['GEO_X'], df_has_no_claim_policy_not_in_swath['GEO_Y'])]
#     # Convert to GeoDataFrame
#     gdf_has_no_claim_policy_not_in_swath = gpd.GeoDataFrame(df_has_no_claim_policy_not_in_swath, geometry=geometry, crs="EPSG:4326") 
#     gdf_has_no_claim_policy_not_in_swath = get_nearest_distance_polygon(gdf_has_no_claim_policy_not_in_swath, gdf)    
#     gdf_has_no_claim_policy_not_in_swath_at_8k = gdf_has_no_claim_policy_not_in_swath[gdf_has_no_claim_policy_not_in_swath['distance'] < 8000]
#     return gdf_has_no_claim_policy_not_in_swath, gdf_has_no_claim_policy_not_in_swath_at_8k

def get_policy_outside_swath (gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, df_policy, gdf_has_no_claim_in_hailswath):
    cat_zipcode  = set(list(gdf_has_claim_inside_swath['ZIP5'].unique()) + list(gdf_has_claim_not_in_swath['ZIP5'].unique()) )
    df_policy_in_zipcode = df_policy[df_policy['ZIP5'].isin(cat_zipcode) ]
    policy_no_in_swath = set(list(gdf_has_claim_inside_swath['POLICY_NO'].unique()) + list(gdf_has_no_claim_in_hailswath['POLICY_NO'].unique()) )
    df_policy_outside_swath = df_policy_in_zipcode[~df_policy_in_zipcode['POLICY_NO'].isin(policy_no_in_swath) ]
    df_has_no_claim_policy_not_in_swath = df_policy_outside_swath[~df_policy_outside_swath['POLICY_NO'].isin(gdf_has_claim_not_in_swath['POLICY_NO'])]
    return df_policy_outside_swath, df_has_no_claim_policy_not_in_swath

def convert_df_to_gdf(df_has_no_claim_policy_not_in_swath):
    geometry = [Point(xy) for xy in zip(df_has_no_claim_policy_not_in_swath['GEO_X'], df_has_no_claim_policy_not_in_swath['GEO_Y'])]
    gdf_has_no_claim_policy_not_in_swath = gpd.GeoDataFrame(df_has_no_claim_policy_not_in_swath, geometry=geometry, crs="EPSG:4326") 
    return gdf_has_no_claim_policy_not_in_swath

def get_outside_swath_nearest_polygon(df_has_no_claim_policy_not_in_swath, gdf):
    gdf_has_no_claim_policy_not_in_swath = convert_df_to_gdf(df_has_no_claim_policy_not_in_swath) 
    gdf_has_no_claim_policy_not_in_swath = get_nearest_distance_polygon(gdf_has_no_claim_policy_not_in_swath, gdf)    
    gdf_has_no_claim_policy_not_in_swath_at_8k = gdf_has_no_claim_policy_not_in_swath[gdf_has_no_claim_policy_not_in_swath['distance'] < 8000]
    return gdf_has_no_claim_policy_not_in_swath, gdf_has_no_claim_policy_not_in_swath_at_8k

def get_summary_level_data(cat_no, cat_date,gdf, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath
        ,df_policy_outside_swath , gdf_has_no_claim_policy_not_in_swath_at_8k
    ):
    df_cat_summary = pd.DataFrame( columns = ['cat', 'cat_date', 'total_shape_count', 'total_Shape_Area',
        'avg_Shape_Area', 'avg_hail_diameter_in', 'max_hail_diameter_in',
        'min_hail_diameter_in', 'std_hail_diameter_in',
        'total_hail_diameter_shape_area', 'avg_hail_diameter_shape_area',
        'avg_fill_opacity', 'total_claim_count_inside_swath',
        'total_policy_count_inside_swath', 'total_claim_count_outside_swatch',
        'total_claim_count_outside_swatch_distance_50_percentile',
        'total_claim_count_outside_swatch_distance_75_percentile',
        'total_claim_count_outside_swatch_distance_90_percentile',
        'total_policy_count_outside_swath_at_8k',
        'total_policy_count_outside_swath'
        ]
        )
    dict_summary = {
        'cat': cat_no,
        'cat_date':cat_date,        
        'total_shape_count': len(gdf),
        'total_Shape_Area' : gdf['Shape_Area'].sum(),
        'avg_Shape_Area': gdf['Shape_Area'].mean(),
        'avg_hail_diameter_in':gdf['diameter_in'].mean(),
        'max_hail_diameter_in':gdf['diameter_in'].max(),
        'min_hail_diameter_in': gdf['diameter_in'].min(),
        'std_hail_diameter_in': gdf['diameter_in'].std(),
        'total_hail_diameter_shape_area': (gdf['diameter_in']*gdf['Shape_Area']).sum(),
        'avg_hail_diameter_shape_area': (gdf['diameter_in']*gdf['Shape_Area']).mean(), 
        'avg_fill_opacity':gdf['fill_opacity'].mean(),
        'total_claim_count_inside_swath': len(gdf_has_claim_inside_swath),    #this will not available, preditor 
        'total_policy_count_inside_swath': len(gdf_has_claim_inside_swath) + len(gdf_has_no_claim_in_hailswath) , 
        'total_claim_count_outside_swath': len(gdf_has_claim_not_in_swath),  #this will not available  predictor
        'total_claim_count_outside_swath_distance_50_percentile': gdf_has_claim_not_in_swath['distance'].quantile(0.5),
        'total_claim_count_outside_swath_distance_75_percentile': gdf_has_claim_not_in_swath['distance'].quantile(0.75),
        'total_claim_count_outside_swath_distance_90_percentile': gdf_has_claim_not_in_swath['distance'].quantile(0.9),
        'total_policy_count_outside_swath_at_8k': len(gdf_has_no_claim_policy_not_in_swath_at_8k),
        'total_policy_count_outside_swath': len(df_policy_outside_swath)  
        }

    df_cat_summary.loc[len(df_cat_summary)] = dict_summary.values()
    return df_cat_summary

def save_raw_gdf (cat_no,cat_date,**kargs):  #use name : gd format
    for  n, gd in kargs.items():
        if n == 'gdf':
            gd.to_file(f'{data_folder}/{cat_no}_{cat_date}_hail_swath__all.gpkg', driver='GPKG')
        elif n == 'summary':
            gd.to_csv(f"{data_folder}/{cat_no}_{cat_date}_summary.csv", index=False)
        else:
            gd['geometry_wkt'] = gd['geometry'].apply(lambda p: p.wkt)
            df_no_geo = pd.DataFrame(gd.drop(columns='geometry'))
            df_no_geo.to_csv(f"{data_folder}/{cat_no}_{cat_date}_{n}.csv", index=False)
    return

def save_raw_gdf_test (cat_no,cat_date,**kargs):  #use name : gd format
    '''
    This is use to save to a specific test folder, not to mix with other data
    '''
    for  n, gd in kargs.items():
        if n == 'gdf':
            gd.to_file(f'{data_folder}/{cat_no}_{cat_date}_hail_swath__all.gpkg', driver='GPKG')
        elif n == 'summary':
            gd.to_csv(f"{data_folder}/{cat_no}_{cat_date}_summary.csv", index=False)
        else:
            gd['geometry_wkt'] = gd['geometry'].apply(lambda p: p.wkt)
            df_no_geo = pd.DataFrame(gd.drop(columns='geometry'))
            df_no_geo.to_csv(f"{data_folder}/{cat_no}_{cat_date}_{n}.csv", index=False)
    return

def stratified_sample_fixed_n(df, strata_cols, n_samples, random_state=None):
    """
    Performs stratified sampling on a pandas DataFrame based on multiple columns
    to achieve a fixed total number of samples.

    :param df: The input DataFrame.
    :param strata_cols: A list of column names to stratify by.
    :param n_samples: The fixed total number of samples to retrieve.
    :param random_state: Seed for the random number generator (for reproducibility).
    :return: A new DataFrame with the stratified sample.
    """
    # 1. Calculate the proportion of each stratum in the original data
    strata_counts = df.groupby(strata_cols).size()
    strata_proportions = strata_counts / len(df)
    
    # 2. Calculate the number of samples needed from each stratum to meet the total sample size
    # Rounding can cause the total number of samples to be slightly off.
    strata_sample_n = np.round(strata_proportions * n_samples).astype(int)
    
    # Adjust to ensure the total sample size matches exactly n_samples
    total_sampled = strata_sample_n.sum()
    if total_sampled != n_samples:
        diff = n_samples - total_sampled
        # Adjust the largest strata to make up the difference
        largest_strata = strata_sample_n.nlargest(abs(diff)).index
        strata_sample_n.loc[largest_strata[:abs(diff)]] += np.sign(diff).astype(int)
    
    # Check total samples after adjustment
    assert strata_sample_n.sum() == n_samples

    # 3. Apply the sampling within each group
    # The function below takes the calculated number of samples for each group
    def sample_group(group, n):
        # Ensure we don't try to sample more rows than available in the group
        n_actual = min(len(group), n)
        return group.sample(n=n_actual, random_state=random_state)

    # Convert the Series of counts to a dictionary for easier application
    sample_counts_dict = strata_sample_n.to_dict()

    # Apply the sampling to the original DataFrame using the pre-calculated sample counts
    sampled_df = df.groupby(strata_cols, group_keys=False).apply(
        lambda x: sample_group(x, sample_counts_dict[tuple(x.name) if isinstance(x.name, tuple) else x.name])
    )
    
    return sampled_df.sample(frac=1, random_state=random_state) 


def merge_gdf_data (cat_no, cat_date, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k):
    df_data = pd.DataFrame()

    for df in [gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath,  gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k]:    
        temp_df = pd.DataFrame()
        #join key    
        temp_df['ACCOUNT_NO'] = df['ACCOUNT_NO']
        temp_df['POLICY_NO'] = df['POLICY_NO']
        #cat info
        temp_df['LOSS_DT'] = cat_date
        temp_df['CAT_NO'] = cat_no
        #geo info
        temp_df['RES_PREM_ADDRESS'] = df['RES_PREM_ADDRESS']
        temp_df['CITY'] = df['CITY']    
        temp_df['STATE'] = df['STATE']
        temp_df['ZIP5'] = df['ZIP5']
        temp_df['GEO_X'] = df['GEO_X']
        temp_df['GEO_Y'] = df['GEO_Y']

        if (df is gdf_has_claim_inside_swath) or (df is gdf_has_no_claim_in_hailswath):
                temp_df['diameter_in'] = df['diameter_in']
        elif (df is gdf_has_claim_not_in_swath) or (df is gdf_has_no_claim_policy_not_in_swath_at_8k):
            temp_df['diameter_in'] = df['nearest_polygon_diameter_in']

        #group
        if (df is gdf_has_no_claim_in_hailswath) or (df is gdf_has_no_claim_policy_not_in_swath_at_8k):
            temp_df['TYPE'] = 'No_Claim_in_HailSwath' if df is gdf_has_no_claim_in_hailswath else 'No_Claim_outside_HailSwath'
            temp_df['Claim_Status'] = 'No_Claim'
        else:
            temp_df['TYPE'] = 'Claim_in_HailSwath' if df is gdf_has_claim_inside_swath else 'Claim_outside_HailSwath'
            temp_df['Claim_Status'] = 'Has_Claim'

        df_data = pd.concat([df_data, temp_df], ignore_index=True)
    return df_data

def prep_model_data(cat_no, cat_date, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k): 
    '''
    this is for training to get balanced dataset,  use stratified sample method to downsample
    '''    
    def gdf_drop_duplicate(gdf):
        gdf.sort_values(by=['ACCOUNT_NO', 'POLICY_NO','diameter_in'], ascending=False, inplace=True)
        gdf.drop_duplicates(subset=['ACCOUNT_NO', 'POLICY_NO'], keep='first', inplace=True)
        return 
    
    for g in [gdf_has_claim_inside_swath,  gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k]:
        g = g.copy()
        #print(len(g))

    for g in [gdf_has_claim_inside_swath,  gdf_has_no_claim_in_hailswath]:
        gdf_drop_duplicate(g)
        
    #under sampling stratify by hail size 
    gdf_has_no_claim_in_hailswath = stratified_sample_fixed_n(gdf_has_no_claim_in_hailswath, ['diameter_in', 'ZIP5'], len(gdf_has_claim_inside_swath), random_state=42)
    #print(len(gdf_has_no_claim_in_hailswath))
    gdf_has_no_claim_policy_not_in_swath_at_8k = stratified_sample_fixed_n(gdf_has_no_claim_policy_not_in_swath_at_8k, ['nearest_polygon_diameter_in', 'ZIP5'], len(gdf_has_claim_not_in_swath), random_state=42)
    #print(len(gdf_has_no_claim_policy_not_in_swath_at_8k))
    
    #merge data
    df_data = merge_gdf_data(cat_no, cat_date,  gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    
    return df_data

def prep_model_data_full(cat_no, cat_date, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k): 
    '''
    this is for actual data for test and predicton to get all dataset,  not use stratified sample method to downsample
    '''  
    def gdf_drop_duplicate(gdf):
        gdf.sort_values(by=['ACCOUNT_NO', 'POLICY_NO','diameter_in'], ascending=False, inplace=True)
        gdf.drop_duplicates(subset=['ACCOUNT_NO', 'POLICY_NO'], keep='first', inplace=True)
        return 
    
    for g in [gdf_has_claim_inside_swath,  gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k]:
        g = g.copy()
        #print(len(g))

    for g in [gdf_has_claim_inside_swath,  gdf_has_no_claim_in_hailswath]:
        gdf_drop_duplicate(g)
        
    #under sampling stratify by hail size 
    #gdf_has_no_claim_in_hailswath = stratified_sample_fixed_n(gdf_has_no_claim_in_hailswath, ['diameter_in', 'ZIP5'], len(gdf_has_claim_inside_swath), random_state=42)
    #print(len(gdf_has_no_claim_in_hailswath))
    #gdf_has_no_claim_policy_not_in_swath_at_8k = stratified_sample_fixed_n(gdf_has_no_claim_policy_not_in_swath_at_8k, ['nearest_polygon_diameter_in', 'ZIP5'], len(gdf_has_claim_not_in_swath), random_state=42)
    #print(len(gdf_has_no_claim_policy_not_in_swath_at_8k))
    
    #merge data
    df_data = merge_gdf_data(cat_no, cat_date,  gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    
    return df_data


def get_claim_history(cat_date):
    sql_claim_history = f'''
    declare @lossdate date = '{cat_date}'
    drop table if exists #tmp_account
    select 
    ACCOUNT_NO,
    Policy_No,
    POL_1ST_INS_DATE,
    GEO_X, 
    GEO_Y
    into #tmp_account
    from POLICY_HISTORY_VIEW b
    where 
    b.INFORCE_BEGIN_DT <= @lossdate AND b.INFORCE_END_DT >@lossdate

    --join with claim table to get claim filed

    select 
    ACCOUNT_NO, 
    count(*) as claim_count,
    sum(COVA) as claim_amount
    from #tmp_account a
    join CLAIMS_HO_HISTPOLICY b on  b.POL_1ST_INS_DATE = a.POL_1ST_INS_DATE and a.GEO_X = b.GEO_X and a.GEO_Y = b.GEO_Y
    where b.LOB = 'Home'
    and b.COV = 'HAIL'
    and b.LOSS_DT < @lossdate
    group by ACCOUNT_NO
    '''
    df_claim_history = gis_query(sql_claim_history)
    return df_claim_history
   

def get_pc_roof_data(df_data, cat_date):
    df_data1 = df_data.copy()
    pc_all_policy = f'''
    select 
    ((SUBSTRING(b.policynumber,1,1) + SUBSTRING(b.policynumber,5,9)))as PolicyNumber,
    c.RoofYearExt,
    b.periodstart,
    b.periodend
    from pc_policy  a  
    join pc_policyperiod b on a.id = b.PolicyID
    join pcx_dwelling_hoe  c on b.id = c.BranchID
    where  a.ProductCode = 'Homeowners' 
    and 
    b.periodstart <= '{cat_date}' and b.periodend > '{cat_date}'
    and ((SUBSTRING(b.policynumber,1,1) + SUBSTRING(b.policynumber,5,9))) is not null
    and RoofYearExt is not null
    '''
    df_all_policy_roof =  pc_query(pc_all_policy)
    df_all_policy_roof = df_all_policy_roof.drop_duplicates()
    df_all_policy_roof.rename(columns={'PolicyNumber': 'POLICY_NO'}, inplace=True)
    df_data1 = df_data1.merge(df_all_policy_roof, on='POLICY_NO', how='left')
    df_data1 = df_data1[pd.to_datetime(df_data1['LOSS_DT']).dt.year >= df_data1['RoofYearExt']]
    #get row number ranked by policy no, roof year desc 
    df_data1 = df_data1.sort_values(by=['POLICY_NO', 'RoofYearExt'], ascending=[True, False])
    df_data1['rank'] = df_data1.groupby('POLICY_NO').cumcount()+1
    #df_data1[df_data1['POLICY_NO'] == '6203421329']
    df_data1 = df_data1[df_data1['rank'] == 1]
    df_data1 = df_data1.drop(columns=['rank', 'periodstart', 'periodend'])
    #join back due some record not found in the database
    df_data = df_data.merge(df_data1[['POLICY_NO', 'RoofYearExt']], on='POLICY_NO', how='left') 
    return df_data


async def get_hail_storm_data(cat_no, cat_date, save_gdf = False):
    '''
    get training balanced dataset
    df = pd.DataFrame(columns=['Account_No', 'Policy_No', 'Loss_Dt', 'Cat_No', 'Address',
       'City', 'State', 'Zip5', 'Geo_X', 'Geo_Y', 'Hail_Diameter_In', 'Type',
       'Claim_Status', 'Wind_Speed', 'Historic_Claim_Count', 'Claim_Amount','Roof_Year'])
    for index, row in df_hail_storm_selected.iterrows():    
        cat_no = row['cat_no']
        cat_date = str(row['LOSS_DT'])[:10]
        print(f"Processing cat_no: {cat_no}, cat_date: {cat_date}")
        df_data = await get_hail_storm_data(cat_no, cat_date, save_gdf = True)
        df_data.columns = df.columns
        df = pd.concat([df, df_data], ignore_index=True)
    df.head()
    
    '''
    gdf = get_hail_storm_polygon(cat_no, cat_date, save_gdf = save_gdf)
    #get all claim related with the cat
    df_claim = get_claim_data(cat_no, cat_date)
    #get all policy span over cat date
    df_policy = get_all_policies_span_over_date(cat_date)
    #get spatial join between polygon and claim
    gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath = get_cat_geodataframe (gdf, df_claim, df_policy)
    gdf_has_claim_not_in_swath = get_nearest_distance_polygon(gdf_has_claim_not_in_swath, gdf)
    #policy outside the swath
    df_policy_outside_swath, df_has_no_claim_policy_not_in_swath = get_policy_outside_swath (gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, df_policy, gdf_has_no_claim_in_hailswath)
    #get nearest polygon for policy outside swath
    gdf_has_no_claim_policy_not_in_swath, gdf_has_no_claim_policy_not_in_swath_at_8k = get_outside_swath_nearest_polygon(df_has_no_claim_policy_not_in_swath, gdf)
    #get summary level data
    df_cat_summary = get_summary_level_data(cat_no, cat_date,gdf, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, df_policy_outside_swath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    #save raw dataset
    dict_gdf = {
        'gdf': gdf,
        'summary': df_cat_summary, 
        'gdf_has_claim_inside_swath':gdf_has_claim_inside_swath, 
        'gdf_has_claim_not_in_swath':gdf_has_claim_not_in_swath, 
        'gdf_has_no_claim_in_hailswath':gdf_has_no_claim_in_hailswath,
        'gdf_has_no_claim_policy_not_in_swath_at_8k': gdf_has_no_claim_policy_not_in_swath_at_8k
    }
    save_raw_gdf (cat_no, cat_date, **dict_gdf)
    #combine all data for model, downsample negative class
    df_data = prep_model_data(cat_no, cat_date, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    #add wind data
    #df_data['wind_speed'] = df_data.apply(lambda row: get_wind_data(row['GEO_Y'], row['GEO_X'], row['LOSS_DT'], row['LOSS_DT'], row['LOSS_DT']), axis=1)
    df_data = await get_wind_dataframe(df_data)
    #add claim history
    df_claim_history = get_claim_history(cat_date)
    df_data = df_data.merge(df_claim_history, on='ACCOUNT_NO', how='left')
    #add pc roof data
    df_data = get_pc_roof_data(df_data, cat_date)
    return df_data


async def get_hail_storm_data_full(cat_no, cat_date, save_gdf = False):
    '''
    get full unbalanced dataset
    df = pd.DataFrame(columns=['Account_No', 'Policy_No', 'Loss_Dt', 'Cat_No', 'Address',
       'City', 'State', 'Zip5', 'Geo_X', 'Geo_Y', 'Hail_Diameter_In', 'Type',
       'Claim_Status', 'Wind_Speed', 'Historic_Claim_Count', 'Claim_Amount','Roof_Year'])
    for index, row in df_hail_storm_selected.iterrows():    
        cat_no = row['cat_no']
        cat_date = str(row['LOSS_DT'])[:10]
        print(f"Processing cat_no: {cat_no}, cat_date: {cat_date}")
        df_data = await get_hail_storm_data_full(cat_no, cat_date, save_gdf = True)
        df_data.columns = df.columns
        df = pd.concat([df, df_data], ignore_index=True)
    df.head()
    
    '''
    gdf = get_hail_storm_polygon(cat_no, cat_date, save_gdf = save_gdf)
    #get all claim related with the cat
    df_claim = get_claim_data(cat_no, cat_date)
    #get all policy span over cat date
    df_policy = get_all_policies_span_over_date(cat_date)
    #get spatial join between polygon and claim
    gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath = get_cat_geodataframe (gdf, df_claim, df_policy)
    gdf_has_claim_not_in_swath = get_nearest_distance_polygon(gdf_has_claim_not_in_swath, gdf)
    #policy outside the swath
    df_policy_outside_swath, df_has_no_claim_policy_not_in_swath = get_policy_outside_swath (gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, df_policy, gdf_has_no_claim_in_hailswath)
    #get nearest polygon for policy outside swath
    gdf_has_no_claim_policy_not_in_swath, gdf_has_no_claim_policy_not_in_swath_at_8k = get_outside_swath_nearest_polygon(df_has_no_claim_policy_not_in_swath, gdf)
    #get summary level data
    df_cat_summary = get_summary_level_data(cat_no, cat_date,gdf, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, df_policy_outside_swath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    #save raw dataset
    dict_gdf = {
        'gdf': gdf,
        'summary': df_cat_summary, 
        'gdf_has_claim_inside_swath':gdf_has_claim_inside_swath, 
        'gdf_has_claim_not_in_swath':gdf_has_claim_not_in_swath, 
        'gdf_has_no_claim_in_hailswath':gdf_has_no_claim_in_hailswath,
        'gdf_has_no_claim_policy_not_in_swath_at_8k': gdf_has_no_claim_policy_not_in_swath_at_8k
    }
    save_raw_gdf_test (cat_no, cat_date, **dict_gdf)
    #combine all data for model, downsample negative class
    df_data = prep_model_data_full(cat_no, cat_date, gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath, gdf_has_no_claim_policy_not_in_swath_at_8k)
    #add wind data
    #df_data['wind_speed'] = df_data.apply(lambda row: get_wind_data(row['GEO_Y'], row['GEO_X'], row['LOSS_DT'], row['LOSS_DT'], row['LOSS_DT']), axis=1)
    df_data = await get_wind_dataframe(df_data)
    #add claim history
    df_claim_history = get_claim_history(cat_date)
    df_data = df_data.merge(df_claim_history, on='ACCOUNT_NO', how='left')
    #add pc roof data
    df_data = get_pc_roof_data(df_data, cat_date)
    return df_data





__all__ = ['monitor_pipeline', 'update_microstrategy', 'get_incremental_update_param']

from ._imports import *
from ._pconfig import *
from ._misc import *
from ._train import get_run_months

def monitor_pipeline(df, model_config, env='dev', logger=None):
    """
    Monitor pipeline that detects target and feature drifting.

    Parameters
    ----------
    df : pd.DataFrame
        Scoring dataset.
    model_config : dict
        Dictionary that stores the model's metadata.
    env : str
        The run environment, 'dev' or 'prod'.
    logger : logging.RootLogger
        Logger for information, warnings, and errors.

    Returns
    -------
    dict
        Dictionary of model monitoring results.
    """

    model_version  = get_model_version(env)
    s3_file        = os.path.join(pconfig[f's3_model_assets_folder_{env}'], f'cancel_v{model_version}.csv')
    kwargs         = {'na_values': pd._libs.parsers.STR_NA_VALUES - {'N/A'}, 'keep_default_na': False}
    reference      = s3_download_and_read_csv(s3_file, read_csv_kwargs=kwargs, env=env)

    concept_drift  = target_drift(df, reference)
    data_drift     = feature_drift(df, reference, model_config)
    null_drift     = feature_null_drift(df, reference, model_config)
    unseen_drift   = feature_unseen_category(df, reference, model_config)

    trackers = {
        'records':           len(df),
        'distribution': {
            'concept_drift': concept_drift,
            'data_drift':    data_drift,
        },
        'null_drift':        null_drift,
        'unseen_drift':      unseen_drift,
    }
    alert = alert_drift(trackers, model_config)
    trackers['alert'] = alert
    return trackers

def target_drift(df, reference):
    drift_preds = chi_square(reference['PREDICTION'], df['PREDICTION'])
    drift_probs = population_stability_index(reference['PROBABILITY'], df['PROBABILITY'])
    return {'PREDICTION': drift_preds, 'PROBABILITY': drift_probs}

def feature_drift(df, reference, model_config):
    df, reference = df.dropna(), reference.dropna()
    drift_feats = {}
    for f in model_config['features']:
        if pd.api.types.is_numeric_dtype(reference[f]):
            drift = population_stability_index(reference[f], df[f])
        else:
            drift = chi_square(reference[f], df[f])
        drift_feats[f] = drift
    return drift_feats

def feature_null_drift(df, reference, model_config):
    null_drift = []
    for f in model_config['features']:
        drift = chi_square(reference[f].isna(), df[f].isna())
        if drift: null_drift.append(f)
    return null_drift

def feature_unseen_category(df, reference, model_config):
    unseen_feats = {}
    for f in model_config['features']:
        if not pd.api.types.is_numeric_dtype(reference[f]):
            unseen = set(df[f].unique()) - set(reference[f].unique())
            if unseen:
                unseen_feats[f] = unseen
    return list(unseen_feats)

def chi_square(ref, new, alpha=0.05):
    categories     = set(ref.unique()).union(set(new.unique()))
    ref_proportion = [(ref==c).sum()/len(ref) for c in categories]
    new_proportion = [(new==c).sum()/len(new) for c in categories]
    r = scipy.stats.chisquare(new_proportion, f_exp=ref_proportion)
    return int(r.pvalue < alpha)

def population_stability_index(ref, new, buckets=10, psi_cutoff=0.2):
    '''
    Calculate the PSI (population stability index)
    Adapted from https://github.com/mwburke/population-stability-index/blob/master/psi.py
    '''

    def sub_psi(e_pct, a_pct):
        if a_pct == 0: a_pct = 0.0001
        if e_pct == 0: e_pct = 0.0001
        return (e_pct - a_pct) * np.log(e_pct / a_pct)

    breaks  = np.linspace(np.min(ref), np.max(ref), buckets+1)
    ref_pct = np.histogram(ref, breaks)[0] / len(ref)
    new_pct = np.histogram(new, breaks)[0] / len(new)

    psi_value = np.sum(sub_psi(ref_pct[i], new_pct[i]) for i in range(0, len(ref_pct)))
    return int(psi_value > psi_cutoff)

def alert_drift(trackers, model_config, k=5):
    alert = False
    if 'error' in trackers:
        if trackers['error'] == 1:
            return False
    if sum(trackers['distribution']['concept_drift'].values()) > 0:
        return True

    model_config['importances'] = {
       k:v for (k,v) in model_config['importances'].items() if k not in ("STATE", "STATE_TERRITORY", "SOURCE_OF_BUSINESS")
    }
    if 'importances' in model_config:
        importances_v = model_config['importances']
        importances_f = sorted(importances_v.items(), key=lambda x: x[1], reverse=True)
        for f in importances_f[:k]:
            if trackers['distribution']['data_drift'][f[0]] > 0:
                alert = True
    return alert

def update_microstrategy(env='prod', logger=None, **incremental_kwargs):
    '''
    currently, we run this function after score pipeline evaluation pipeline
    all dataset will be updated although evaluatin pipeline only impact eval and impact
    when score month does not exist in eval and impact in incremental mode, it will create empty df. merge result will be the data from s3
    if want to update all record
    update_microstrategy(env, logger)
    if just want to increment load what have been changed
    incremental_kwargs = get_incremental_update_param(eval_month, eval_month) #begin and end to update a few month
    update_microstrategy(env=env, logger=logger, **incremental_kwargs)
    '''
    error_status = 0
    try:        
        df = pd.DataFrame()
        eds_file_list = [i for i in s3_list_folder_contents(pconfig[f's3_model_output_eds_folder_{env}'], env=env) if ".csv" in i]
        if incremental_kwargs:
            eds_file_list = [i for i in eds_file_list if (i.split('/')[-1].split('_')[1] in incremental_kwargs['target_month_first_day']) ]  #
        for f in eds_file_list:        
            t = s3_download_and_read_csv(f, env=env)
            t['filename'] = f
            t['runstart'] = f.split('_')[-3]
            t['runend'] = pd.to_datetime(t['runstart']).apply(lambda x:DtFind.last_day_of_month(x)).astype(str)
            t['rundate'] = f.split('_')[-2]
            t['timestamp'] = dt.datetime.strptime(f.split('_')[-1].replace('.csv',''),'%Y%m%d%H%M%S').strftime("%Y-%m-%d %H:%M:%S")
            if df.shape[0] == 0:
                df = t
            else:
                df = pd.concat([df,t])        
        df = df.sort_values(by=['rundate','timestamp'], ascending=False )
        df = df.drop_duplicates(subset=['POLICY_NO', 'rundate'], keep='first')
        df_sum = df.groupby(['filename','rundate','runstart','runend','timestamp']).agg(
            policy_count=('POLICY_NO','count'),
            prediction_average=('PREDICTION','mean'),
            prediction_min=('PREDICTION','min'),
            prediction_max=('PREDICTION','max'),
            prediction_std=('PREDICTION','std')
        ).reset_index()
        if logger:
            logger.info(f"Mstr df_sum: {len(df_sum)}.")
        
        monitor_files = s3_list_folder_contents(
                os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'monitor/'),
                pattern=r'^.*cancel_.*\.json$',
                env=env,
            )
        if incremental_kwargs:
            monitor_files  = [i for i in monitor_files if  (i.split('/')[-1].split('_')[1] in incremental_kwargs['run_dt']) ]  #
        concept_drift_df = pd.DataFrame(columns=['Date','Prediction','Probability'])
        for j in monitor_files:
            file_path  = os.path.join(config['local']['local_folder'], os.path.basename(j))
            s3_download_file(j, file_path, env=env)
            with open(file_path,"r") as f:
                dtime = j.split('/')[-1].split("_")[1]
                data = json.load(f)
                if 'distribution' in data.keys():
                    pred_drift = data['distribution']['concept_drift']['PREDICTION']
                    prob_drift = data['distribution']['concept_drift']['PROBABILITY']
                    t = pd.DataFrame.from_dict({'Date':[dtime],'Prediction':[pred_drift],'Probability':[prob_drift]})
                    concept_drift_df = pd.concat([concept_drift_df,t])
            
            os.remove(file_path)
        concept_drift_df = concept_drift_df.drop_duplicates()

        data_drift_df = pd.DataFrame(columns=['Date','Feature','Drift'])
        for j in monitor_files:
            file_path  = os.path.join(config['local']['local_folder'], os.path.basename(j))
            s3_download_file(j, file_path, env=env)
            with open(file_path,"rb") as f:
                dtime = j.split('/')[-1].split("_")[1]
                data = json.load(f)
                if 'distribution' in data.keys():                
                    for k, v in data['distribution']['data_drift'].items():                
                        t = pd.DataFrame.from_dict({'Date':[dtime],'Feature':[k],'Drift':[v]})
                        data_drift_df = pd.concat([data_drift_df,t])            
            os.remove(file_path)
        data_drift_df = data_drift_df.drop_duplicates()

        evaluate_files = s3_list_folder_contents(
                os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'evaluate/'), 
                pattern=r'^.*cancel_.*\.json$',
                env=env,
            )
        if incremental_kwargs:
            evaluate_files  = [i for i in evaluate_files if  (i.split('/')[-1].split('_')[2] in incremental_kwargs['target_month_first_day']) ]  #        
        eval_df = pd.DataFrame(columns=['Date','Metric','Value'])
        for j in evaluate_files:
            file_path = os.path.join(config['local']['local_folder'], os.path.basename(j))
            s3_download_file(j, file_path, env=env)
            with open(file_path,"rb") as f:
                dtime = j.split('/')[-1].split("_")[2]
                data = json.load(f)
                if 'all' in data.keys():
                    for k, v in data['all'].items():
                        t = pd.DataFrame.from_dict({'Date':[dtime],'Metric':[k],'Value':[v]})
                        eval_df = pd.concat([eval_df,t])
            os.remove(file_path)
        eval_df = eval_df.drop_duplicates()
        eval_df['Value'] = eval_df['Value'].astype(float)

        impact_files = s3_list_folder_contents(
                os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'impact/'),
                pattern=r'^.*cancel_.*\.json$',
                env=env,
            )
        if incremental_kwargs:
            impact_files  = [i for i in impact_files if  (i.split('/')[-1].split('_')[2] in incremental_kwargs['target_month_first_day']) ]  #
        
        impact_df = pd.DataFrame(columns=['Date','Metrics','Values','Datetime'])
        
        if impact_files:   #when there are files
            for j in impact_files:
                file_path  = os.path.join(config['local']['local_folder'], os.path.basename(j))
                s3_download_file(j, file_path, env=env)
                with open(file_path,"rb") as f:
                    date = j.split('/')[-1].split("_")[2]
                    datetime = j.split('/')[-1].split("_")[-1][:8]        
                    data = json.load(f)
                    for k0 in data.keys():                
                        for k, v in data[k0].items():                
                            t = pd.DataFrame.from_dict({'Date':[date],'Metrics':[k0+': '+k],'Values':[v], 'Datetime':[datetime]})
                            impact_df = pd.concat([impact_df,t])            
                os.remove(file_path)
            impact_df = impact_df.sort_values(by='Datetime').drop_duplicates(subset = ['Date','Metrics'], keep='last')
            impact_df[['Category', 'Metrics']] = impact_df['Metrics'].str.split(': ', n=1, expand=True)
            impact_df = impact_df[['Date', 'Category', 'Metrics', 'Values', 'Datetime']]
            impact_df['Values'] = impact_df['Values'].astype(float)
            impact_df = impact_df.pivot(index=['Category', 'Date',], columns=['Metrics'], values='Values').reset_index()
        else: # empty impact_files, this create empty df to merge with 
            impact_df = pd.DataFrame(columns = ['Category', 'Date', 'accounts predicted cancel',
            'accounts predicted cancel & called', 'accounts retained',
            'average call duration (sec)', 'calls inbound', 'calls outbound',
            'calls total', 'premium gained'])

        if incremental_kwargs:
            df_sum, concept_drift_df, data_drift_df, eval_df, impact_df = upsert_all_evaluation(df_sum, concept_drift_df, data_drift_df, eval_df, impact_df, env)

        
        #sort df
        df_sum.sort_values(by ='rundate', inplace=True)
        concept_drift_df.sort_values(by ='Date', inplace=True)
        data_drift_df.sort_values(by ='Date', inplace=True)
        eval_df.sort_values(by ='Date', inplace=True)
        impact_df.sort_values(by =['Category','Date'], inplace=True)

        save_output_mstr(df_sum, concept_drift_df, data_drift_df, eval_df, impact_df, env, athena=True)

    except Exception as e:
        if logger:
            logger.error('Error', exc_info=e)
        error_status = 1
    finally:
        return error_status
        
    
def save_output_mstr(df_sum, concept_drift_df, data_drift_df, eval_df, impact_df, env, athena=False):
    s3_save_and_upload_file(
        df_sum, 
        s3_file_path=os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'cancel_record_counts.csv'),
        overwrite=True,
        env=env,
    )
    s3_save_and_upload_file(
        concept_drift_df, 
        s3_file_path=os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'concept_drift.csv'),
        overwrite=True,
        env=env,
    )
    s3_save_and_upload_file(
        data_drift_df, 
        s3_file_path=os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'data_drift.csv'),
        overwrite=True,
        env=env,
    )
    s3_save_and_upload_file(
        eval_df, 
        s3_file_path=os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'eval.csv'),
        overwrite=True,
        env=env,
    )
    s3_save_and_upload_file(
        impact_df, 
        s3_file_path=os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'impact.csv'),
        overwrite=True,
        env=env,
    )

    if athena:
        s3_save_and_upload_file(
            df_sum, 
            s3_file_path=os.path.join(
                pconfig[f's3_model_output_mstr_folder_{env}'], 
                'Athena/cancel_record_counts/', 
                'cancel_record_counts.parquet'
            ),
            overwrite=True,
            env=env,
        )
        s3_save_and_upload_file(
            concept_drift_df, 
            s3_file_path=os.path.join(
                pconfig[f's3_model_output_mstr_folder_{env}'],
                'Athena/concept_drift/', 
                'concept_drift.parquet'
            ),
            overwrite=True,
            env=env,
        )
        s3_save_and_upload_file(
            data_drift_df, 
            s3_file_path=os.path.join(
                pconfig[f's3_model_output_mstr_folder_{env}'],
                'Athena/data_drift/', 
                'data_drift.parquet'
            ),
            overwrite=True,
            env=env,
        )
        s3_save_and_upload_file(
            eval_df, 
            s3_file_path=os.path.join(
                pconfig[f's3_model_output_mstr_folder_{env}'],
                'Athena/eval/',
                'eval.parquet'
            ),
            overwrite=True,
            env=env,
        )
        s3_save_and_upload_file(
            impact_df, 
            s3_file_path=os.path.join(
                pconfig[f's3_model_output_mstr_folder_{env}'],
                'Athena/impact/',
                'impact.parquet'
            ),
            overwrite=True,
            env=env,
        )

def upsert_all_evaluation(df_sum, concept_drift_df, data_drift_df, eval_df, impact_df, env):
    df_sum = upsert_single_evaluation(df_sum, 
                            'runstart', 
                            os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'cancel_record_counts.csv'),
                            env)

    concept_drift_df = upsert_single_evaluation(concept_drift_df, 
                            'Date', 
                            os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'concept_drift.csv'),
                            env)
    
    data_drift_df = upsert_single_evaluation(data_drift_df, 
                            'Date', 
                            os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'data_drift.csv'),
                            env)

    eval_df = upsert_single_evaluation(eval_df, 
                            'Date', 
                            os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'eval.csv'),
                            env)

    impact_df = upsert_single_evaluation(impact_df, 
                            'Date', 
                            os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'impact.csv'),
                            env)
    return df_sum, concept_drift_df, data_drift_df, eval_df, impact_df

def upsert_single_evaluation(df, join_key, s3_file_path, env):
    df = df.copy()    
    df_aws = s3_download_and_read_file(s3_file_path=s3_file_path, env=env)
    df = pd.concat([df_aws[~df_aws[join_key].isin(df[join_key])], df], ignore_index=True)
    return df

def get_incremental_update_param(month_start, month_end):
    months = get_run_months(month_start, month_end)
    kwargs ={}
    kwargs['target_month'] = months
    run_day=pconfig['run_day']
    if str(run_day).isnumeric():
        kwargs['run_dt']      = [DtFind.first_day_last_month(dt.datetime.strptime(i, '%Y-%m')).replace(day=run_day).strftime('%Y-%m-%d') for i in months]
        kwargs['target_month_first_day'] = [dt.datetime.strptime(i, '%Y-%m').strftime('%Y-%m-%d') for i in months]
    return kwargs


from ._imports import *
from ._config import config
from ._awsbase import Aws_base

class Aws_comprehend (Aws_base):
    @classmethod
    def _get_client(cls, env):
        return super()._get_client(env, 'comprehend')  

    # Function to replace VINs with {VIN}
    @staticmethod
    def replace_vin(text):
        vin_pattern = r'\b[A-HJ-NPR-Z0-9]{17}\b'
        return re.sub(vin_pattern, 'VIN', text)
    
    
    
    @staticmethod
    def dict_to_dataframe(data_dict, prefix=None):
        """
        Convert a dictionary into a pandas DataFrame where the keys are the column names.
        Optionally, add a prefix to the column names.

        Parameters:
        data_dict (dict): The dictionary to convert.
        prefix (str): The prefix to add to the column names (default is None).

        Returns:
        pd.DataFrame: The resulting DataFrame.
        """
        # Convert the dictionary to a DataFrame
        df = pd.DataFrame([data_dict])
        # Add prefix to column names if provided
        if prefix:
            df = df.add_prefix(prefix)
        return df
    
def client(env):
    return Aws_comprehend._get_client(env)

@atexit.register
def comprehend_close():
    for client in Aws_comprehend.clients.values():
        client.close()


def remove_pii (text, env):
    clean_text = Aws_comprehend.replace_vin(text)
    # reversed to not modify the offsets of other entities when substituting
    response = client(env)[env].detect_pii_entities(
    Text= clean_text,
    LanguageCode='en'
    )
    for NER in reversed(response['Entities']):
        clean_text = clean_text[:NER['BeginOffset']] + NER['Type'] + clean_text[NER['EndOffset']:]
    return clean_text

def get_targeted_sentiment (df, env, join_keys=[]):
    targeted_sentiment = pd.DataFrame()
    error_rows = []

    # Reset the index of the DataFrame in place
    df.reset_index(drop=True, inplace=True)

    for i, row in df.iterrows():    
        txt = row['concatenated_messages'].encode('utf-8')[:5000].decode('utf-8', 'ignore')
        if len(row['concatenated_messages'].encode('utf-8')) > 5000:
            message_length_exceeded = 1
        else:
            message_length_exceeded = 0

        target_sentiment_response = client(env)[env].detect_targeted_sentiment(
            Text=txt,
            LanguageCode='en'
        )

        for ent in target_sentiment_response['Entities']:
            comp_sentiment_score = ent['Mentions'][0]['MentionSentiment']['SentimentScore']
            comp_sentiment_score_df = Aws_comprehend.dict_to_dataframe(comp_sentiment_score, 'comp_sentiment_score_')

            comp_mentions = pd.DataFrame()
            for col in ['Score', 'GroupScore', 'Text', 'Type', 'BeginOffset', 'EndOffset']:
                comp_mentions.loc[0, col] = ent['Mentions'][0][col]
            comp_sentiment = pd.DataFrame({'comp_sentiment': [ent['Mentions'][0]['MentionSentiment']['Sentiment']]})
            temp = comp_mentions.join(comp_sentiment).join(comp_sentiment_score_df)
            for j in join_keys:
                temp[j] = row[j]
            temp['message_length_exceeded'] = message_length_exceeded
            temp = temp[join_keys + ['message_length_exceeded'] + list(comp_mentions.columns) + list(comp_sentiment.columns) + list(comp_sentiment_score_df.columns)]

            targeted_sentiment = pd.concat([targeted_sentiment, temp])        

    return targeted_sentiment, error_rows




from __future__ import division
from ._imports import *
from scipy.stats import shapiro,f_oneway, t, norm, ttest_ind, levene
from math import atanh, pow
from numpy import tanh


#get one way anova 
def get_one_way_anova(*arg):    
    '''
    tillamook = [0.0571, 0.0813, 0.0831, 0.0976, 0.0817, 0.0859, 0.0735,
             0.0659, 0.0923, 0.0836]
    newport = [0.0873, 0.0662, 0.0672, 0.0819, 0.0749, 0.0649, 0.0835,
            0.0725]
    petersburg = [0.0974, 0.1352, 0.0817, 0.1016, 0.0968, 0.1064, 0.105]
    magadan = [0.1033, 0.0915, 0.0781, 0.0685, 0.0677, 0.0697, 0.0764,
            0.0689]
    get_one_way_anova(tillamook, newport, petersburg, magadan)
    '''    
    f_value, p_value =  f_oneway(*arg)
    return f_value, p_value

def get_group_t_test(*arg):
    t_value, p_value =  ttest_ind(*arg)
    return t_value, p_value


 
def check_data_normal_distribuation (x):
    '''
    statistic prerequisite

    '''
    p_value = shapiro(x).statistic
    return p_value

def check_data_variance_homogeneousness(*group):
    '''
    statistic prerequisite
    for i in [[s_qualtric_promoters, s_qualtric_Passives, s_qualtric_Detractors], [s_hugg_promoters,s_hugg_Passives, s_hugg_Detractors]]:
        print(check_data_variance_homogeneousness(*i))
    '''
    p_value = levene(*group).statistic
    return p_value



#https://github.com/psinger/CorrelationStats/blob/master/corrstats.py
#this is for compare two correlation if statistically significant
def rz_ci(r, n, conf_level = 0.95):
    zr_se = pow(1/(n - 3), .5)
    moe = norm.ppf(1 - (1 - conf_level)/float(2)) * zr_se
    zu = atanh(r) + moe
    zl = atanh(r) - moe
    return tanh((zl, zu))

def rho_rxy_rxz(rxy, rxz, ryz):
    num = (ryz-1/2.*rxy*rxz)*(1-pow(rxy,2)-pow(rxz,2)-pow(ryz,2))+pow(ryz,3)
    den = (1 - pow(rxy,2)) * (1 - pow(rxz,2))
    return num/float(den)

def dependent_corr(xy, xz, yz, n, twotailed=True, conf_level=0.95, method='steiger'):
    """
    Calculates the statistic significance between two dependent correlation coefficients
    @param xy: correlation coefficient between x and y
    @param xz: correlation coefficient between x and z
    @param yz: correlation coefficient between y and z
    @param n: number of elements in x, y and z
    @param twotailed: whether to calculate a one or two tailed test, only works for 'steiger' method
    @param conf_level: confidence level, only works for 'zou' method
    @param method: defines the method uses, 'steiger' or 'zou'
    @return: t and p-val

    nps_qsent_cor = df_qualtrics_huggingface['nps_score_q'].corr(df_qualtrics_huggingface['Sentiment_q'])
    nps_hsent_cor = df_qualtrics_huggingface['nps_score_q'].corr(df_qualtrics_huggingface['Sentiment_h']) 
    sent_cor =df_qualtrics_huggingface['Sentiment_q'].corr(df_qualtrics_huggingface['Sentiment_h'])
    print(nps_qsent_cor, nps_hsent_cor, sent_cor, len(df_qualtrics_huggingface)) 

    dependent_corr(0.0457171499666752, 0.07611670526658383, 0.5185121141233301, 2576, twotailed=True, method='steiger')

    """
    if method == 'steiger':
        d = xy - xz
        determin = 1 - xy * xy - xz * xz - yz * yz + 2 * xy * xz * yz
        av = (xy + xz)/2
        cube = (1 - yz) * (1 - yz) * (1 - yz)

        t2 = d * np.sqrt((n - 1) * (1 + yz)/(((2 * (n - 1)/(n - 3)) * determin + av * av * cube)))
        p = 1 - t.cdf(abs(t2), n - 3)

        if twotailed:
            p *= 2

        return t2, p
    elif method == 'zou':
        L1 = rz_ci(xy, n, conf_level=conf_level)[0]
        U1 = rz_ci(xy, n, conf_level=conf_level)[1]
        L2 = rz_ci(xz, n, conf_level=conf_level)[0]
        U2 = rz_ci(xz, n, conf_level=conf_level)[1]
        rho_r12_r13 = rho_rxy_rxz(xy, xz, yz)
        lower = xy - xz - pow((pow((xy - L1), 2) + pow((U2 - xz), 2) - 2 * rho_r12_r13 * (xy - L1) * (U2 - xz)), 0.5)
        upper = xy - xz + pow((pow((U1 - xy), 2) + pow((xz - L2), 2) - 2 * rho_r12_r13 * (U1 - xy) * (xz - L2)), 0.5)
        return lower, upper
    else:
        raise Exception('Wrong method!')

def independent_corr(xy, ab, n, n2 = None, twotailed=True, conf_level=0.95, method='fisher'):
    """
    Calculates the statistic significance between two independent correlation coefficients
    @param xy: correlation coefficient between x and y
    @param xz: correlation coefficient between a and b
    @param n: number of elements in xy
    @param n2: number of elements in ab (if distinct from n)
    @param twotailed: whether to calculate a one or two tailed test, only works for 'fisher' method
    @param conf_level: confidence level, only works for 'zou' method
    @param method: defines the method uses, 'fisher' or 'zou'
    @return: z and p-val
    """

    if method == 'fisher':
        xy_z = 0.5 * np.log((1 + xy)/(1 - xy))
        ab_z = 0.5 * np.log((1 + ab)/(1 - ab))
        if n2 is None:
            n2 = n

        se_diff_r = np.sqrt(1/(n - 3) + 1/(n2 - 3))
        diff = xy_z - ab_z
        z = abs(diff / se_diff_r)
        p = (1 - norm.cdf(z))
        if twotailed:
            p *= 2

        return z, p
    elif method == 'zou':
        L1 = rz_ci(xy, n, conf_level=conf_level)[0]
        U1 = rz_ci(xy, n, conf_level=conf_level)[1]
        L2 = rz_ci(ab, n2, conf_level=conf_level)[0]
        U2 = rz_ci(ab, n2, conf_level=conf_level)[1]
        lower = xy - ab - pow((pow((xy - L1), 2) + pow((U2 - ab), 2)), 0.5)
        upper = xy - ab + pow((pow((U1 - xy), 2) + pow((ab - L2), 2)), 0.5)
        return lower, upper
    else:
        raise Exception('Wrong method!')
    



#extract 5 most recent full dataset and save to local full folder
#https://www.google.com/search?q=python+a+wraper+funciton+contain+mutiple+funciton+one+of+them+are+async%2C+shoud+I+change+all+to+async&sca_esv=286534660d0b6d5d&sxsrf=ANbL-n7fW0CbBRJfM3_-4kYhR-v9W0SuXA%3A1772733943664&ei=98WpaeGjKMih5NoPmaao4Q8&biw=2124&bih=1015&ved=0ahUKEwih-_jHrImTAxXIEFkFHRkTKvwQ4dUDCBE&uact=5&oq=python+a+wraper+funciton+contain+mutiple+funciton+one+of+them+are+async%2C+shoud+I+change+all+to+async&gs_lp=Egxnd3Mtd2l6LXNlcnAiZHB5dGhvbiBhIHdyYXBlciBmdW5jaXRvbiBjb250YWluIG11dGlwbGUgZnVuY2l0b24gb25lIG9mIHRoZW0gYXJlIGFzeW5jLCBzaG91ZCBJIGNoYW5nZSBhbGwgdG8gYXN5bmNI3wdQAFjKBXAAeACQAQCYAWugAe8EqgEDNi4xuAEDyAEA-AEBmAIAoAIAmAMA4gMFEgExIECSBwCgB_IdsgcAuAcAwgcAyAcAgAgA&sclient=gws-wiz-serp
import asyncio
from ._imports import *

#function to call api and get json data, handle erro 
async def submit_wind_history_job(lat, lon, start=None, end=None):
    '''
    Submit parameter to gis web API to get wind data and status 
    Parameters
    ----------
    lat : float 
                Latitude.
    lon : float 
                Longitude.
    start1 : str 
                start datetime in yyyy-MM-dd.
    end1 : str 
                end datetime in yyyy-MM-dd.
    Returns
    ----------
    data : dict 
                data returned in json.
    status : str 
                job status.
    Examples
    ----------
    data, status = submit_wind_history_job(lat, lon, '2022-01-01', '2022-11-17')
    '''
    data = dict()
    status = 'error'
    try:
        SERVICE_BASE = "https://gisdev.amica.com/arcgis/rest/services/GPTools/GetWindHistory/GPServer/GetWindHistory"
        url = f"{SERVICE_BASE}/execute"
        params = {
        "f": "json",
        # Adjust param names to match the REST page:
        # Examples shown; rename if your tool uses "latitude/longitude" or "Point" instead.
        "Latitude": lat,
        "Longitude": lon,
        "BeginDT": start,
        "EndDT": end
    }
        r = requests.post(url, data=params, verify=True, timeout=60)  #SSL set False for testing
        r.raise_for_status()
        data = r.json()
        status = 'success'
        return data, status
    except Exception as e:
        print(f"Error occurred: {e}")
        return data, e
    
#get wind event data into dataframe
def get_dataframe_from_wind_data(data, job_status):
    '''
    get wind event data into dataframe
    Parameters
    ----------
    data : dict 
                data returned in json.
    job_status : str 
                job status.
    Returns
    ----------
    df : pd.DataFrame 
                dataframe with wind event data.
    Examples
    ----------
    df = get_dataframe_from_wind_data(data)
    '''
    df = pd.DataFrame()
    if job_status == 'success':
        row = []
        try:
            for event in data['results'][0]['value']['events']:
                timestamp = event.get('date')
                wind_speed = event.get('WindSpeed')    
                row.append({
                    'timestamp': timestamp,
                    'wind_speed': wind_speed        
                })
            df = pd.DataFrame(row)
        except Exception as e:
            print(f"Error occurred while processing data: {e}")
            print(data)
            raise e
    return df

def get_wind_speed_value_from_ref_df(ref_df, search_date):
    '''
    function that use the ref df to get data 
    Parameters
    ----------
    ref_df : pd.DataFrame 
                reference dataframe with wind event data.
    
    search_date : str 
                start datetime in yyyy-MM-dd.    
    Returns
    ----------
    wind_speed_value : int
                wind speed.
    Examples
    ----------
    wind_speed_value = get_wind_speed_value_from_ref_df(ref_df, search_date)
    '''
    if ref_df.empty:
        return None
    wind_speed_value = None
    mask = (ref_df['timestamp'] == search_date)
    filtered_df = ref_df.loc[mask, ['timestamp', 'wind_speed']]
    if not filtered_df.empty:
        wind_speed_value = filtered_df.iloc[0]['wind_speed'] 
    return wind_speed_value

async def get_wind_data(lat, lon, start=None, end=None, search_date='2022-01-29'):
    '''
    wrapper function to get wind data 
    Parameters
    ----------
    lat : float 
                Latitude.
    lon : float 
                Longitude.
    start : str 
                start datetime in yyyy-MM-dd.
    end : str 
                end datetime in yyyy-MM-dd.
    Returns
    ----------
    wind_speed_value : int
                wind speed.
    Examples
    ----------
    wind_speed_value = get_wind_data(lat, lon, '2022-01-01', '2022-11-17', '2022-01-29')
    '''
    if pd.to_datetime(search_date).year <2020:
        return None
    try:
        data, status = await submit_wind_history_job(lat, lon, start=start, end=end)
        await asyncio.sleep(1)
        df = get_dataframe_from_wind_data(data, status)
        wind_speed_value = get_wind_speed_value_from_ref_df(df, search_date)
    except Exception as e:
        #print(f"Error occurred in get_wind_data: {e}")
        wind_speed_value = None

    return wind_speed_value

async def get_wind_dataframe(df):    
    #df['wind_speed'] = await asyncio.gather(*(get_wind_data(x, y, z, z1, z2) for x, y, z, z1, z2 in df[['GEO_Y', 'GEO_X', 'start', 'end', 'search_date']].itertuples(index=False)))
    df['wind_speed'] = await asyncio.gather(*(get_wind_data(x, y, z, z1, z2) for x, y, z, z1, z2 in zip(df['GEO_Y'], df['GEO_X'],  df['LOSS_DT'], df['LOSS_DT'], df['LOSS_DT'])))
    return df



from datetime import date
from ._imports import *


class HailData:

    '''
    hd = HailData()
    data = hd.get_hail_data()
    df, gdf = hd.extract_data_to_df(data)
    hd.save_data_csv(df, gdf, path= r'D:\Cao_A036898\project\hail\gis_file' )
    print(f"hail_swath_{hd.file_timestamp}.gpkg")
    gdf.head()
    
    '''

    TASK_URL = "https://gisdev.amica.com/arcgis/rest/services/GPTools/GetHailMap/GPServer/GetHailMap/"  #submitJob

    DATE_PARAM_NAME = "hailDate"

    def __init__ (self, cat_date = '2015-04-19'):
        self.cat_date = cat_date
        self.file_timestamp = dt.datetime.now().strftime('%Y%m%d_%H%M%S')


    def submit_gp_job(self, task_url, date_param_name, submit_date):
        """Submits an asynchronous geoprocessing job."""
        submit_url = f"{task_url}/submitJob"
        
        # Parameters for the REST request, including the date
        params = {
            date_param_name: submit_date,
            'f': 'json' # Requesting a JSON response
        }
        
        # If using a token, add it to params:
        # params['token'] = TOKEN

        print(f"Submitting job to {submit_url} with date {submit_date}...")
        response = requests.post(submit_url, data=params)
        response.raise_for_status()
        job_info = response.json()
        
        if 'jobId' in job_info:
            job_id = job_info['jobId']
            print(f"Job submitted with ID: {job_id}")
            return job_id
        else:
            raise Exception(f"Failed to submit job: {job_info.get('messages')}")

    def monitor_job_status(self, task_url, job_id):
        """Monitors the job status periodically until completion."""
        job_url = f"{task_url}/jobs/{job_id}"
        
        # If using a token, add it to the URL query parameters
        # job_url = f"{job_url}?token={TOKEN}" 

        status = ""
        while status not in ["esriJobSucceeded", "esriJobFailed", "esriJobCancelled"]:
            print(f"Checking job status (current: {status})...")
            time.sleep(2) # Wait a bit before checking again to avoid overwhelming the server
            
            response = requests.get(job_url, params={'f': 'json'})
            response.raise_for_status()
            job_response = response.json()
            status = job_response.get('jobStatus')
            
            if status == "esriJobFailed":
                print("Job failed. Messages:")
                for msg in job_response.get('messages', []):
                    print(f"- {msg['description']}")
                raise Exception("Geoprocessing job failed.")
                
        print(f"Job finished with status: {status}")
        return job_response

    def get_job_results(self, job_url, job_response):
        """Retrieves the URLs to the job results."""
        results_urls = {}
        if 'results' in job_response:
            for param_name, param_info in job_response['results'].items():
                result_url = f"{job_url}/{param_info['paramUrl']}"
                
                # If using a token, append it to the result URL
                # result_url = f"{result_url}?token={TOKEN}" 
                
                results_urls[param_name] = result_url
        return results_urls

    def get_hail_data(self):
        try:
            # 1. Submit the job
            job_id = self.submit_gp_job(self.TASK_URL, self.DATE_PARAM_NAME, self.cat_date)
            job_url = f"{HailData.TASK_URL}/jobs/{job_id}"

            # 2. Monitor status until finished
            final_job_info = self.monitor_job_status(HailData.TASK_URL, job_id)

            # 3. Get results URLs
            result_urls = self.get_job_results(job_url, final_job_info)
            # print("\nJob results URLs:")
            # for name, url in result_urls.items():
            #     print(f"- {name}: {url}")

            # 4. get async results 
            response = requests.get(url=result_urls['hailSwath'] + "?f=pjson")
            data = response.json()
            return data

        except Exception as e:
            print(f"\nAn error occurred: {e}")

    
    def extract_data_to_df (self, data):        
        features = data['value']['features']
        attrs_rows = []
        geoms = []
        for f in features:
            attrs = (f.get("attributes") or {}).copy()
            geom = f.get("geometry") or {}
            shp = esri_polygon_to_shapely(geom)
            if shp is not None:
                attrs_rows.append(attrs)
                geoms.append(shp)
        df = pd.DataFrame(attrs_rows)
        gdf = gpd.GeoDataFrame(df, geometry=geoms, crs=f"EPSG:{4326}")  #by default use 4326
        return df, gdf
    
    def save_data_csv(self,df, gdf, path= None):

        file_path =  path if path is not None else ''  

        # Geopackage (single-file, recommended)        
        gdf.to_file(os.path.join(file_path, f"hail_swath_{self.file_timestamp}.gpkg"), layer="hail_swath", driver="GPKG")

        # CSV with WKT for systems that aren't GIS-aware
        df_out = df.copy()
        df_out["WKT"] = gdf.geometry.to_wkt()
        df_out.to_csv(os.path.join(file_path, f"hail_swath_{self.file_timestamp}.csv"), index=False)
        return 


# ------------------------------
# 4) ArcGIS rings → Shapely geometry → GeoDataFrame
# ------------------------------
def ring_area(coords):
    # Shoelace formula (signed area); coords are closed rings [ (x,y), ..., (x,y) ]
    a = 0.0
    for i in range(len(coords) - 1):
        x1, y1 = coords[i]
        x2, y2 = coords[i+1]
        a += x1*y2 - x2*y1
    return a / 2.0

def esri_polygon_to_shapely(geom_esri):
    """
    Convert ArcGIS polygon JSON (with 'rings') to Shapely Polygon/MultiPolygon.
    Positive signed area -> outer ring; negative -> hole for the most recent outer.
    """
    rings = geom_esri.get("rings", [])
    if not rings:
        return None

    outers, holes_list = [], []
    for ring in rings:
        if ring[0] != ring[-1]:
            ring = ring + [ring[0]]
        if ring_area(ring) > 0:
            # new outer shell
            outers.append(ring)
            holes_list.append([])
        else:
            # hole goes to the most recent outer
            if holes_list:
                holes_list[-1].append(ring)
            else:
                # fallback: treat as standalone outer if no outer has been seen yet
                outers.append(ring)
                holes_list.append([])

    if len(outers) == 1:
        return Polygon(outers[0], holes=holes_list[0] if holes_list else None)

    # multiple outers → MultiPolygon
    polys = []
    for i, outer in enumerate(outers):
        holes = holes_list[i] if i < len(holes_list) else None
        polys.append(Polygon(outer, holes=holes))
    return MultiPolygon(polys)


    


from ._imports import *

def inspect_module(object):
    '''
    check object/function source
    inspect_module(evaluate_pipeline)
    '''
    import inspect
    return inspect.getmodule(object)

def filter_list (keyword, alist=[]):
    '''
    filter_list('corr', dir())
    '''
    f= keyword
    p = f'.*{f}.*'

    return list(filter(lambda x: re.search(p, x, re.IGNORECASE),   alist ))


from ._imports import *
from ._config import config

class Aws_base:
    clients = {}

    @classmethod
    def _get_client(cls, env, service):    #only dl and dev have permission 
        if env not in cls.clients:
            if config.has_option(f'aws_{env}','aws_session_token'):
                cls.clients[env] = boto3.client(
                    service_name            = service,
                    region_name             = config[f'aws_{env}']['region'], 
                    aws_access_key_id       = config[f'aws_{env}']['aws_access_key_id'],
                    aws_secret_access_key   = config[f'aws_{env}']['aws_secret_access_key'],
                    aws_session_token       = config[f'aws_{env}']['aws_session_token']
                )
            else:
                cls.clients[env] = boto3.client(
                    service_name            = service,
                    region_name             = config[f'aws_{env}']['region'], 
                    aws_access_key_id       = config[f'aws_{env}']['aws_access_key_id'],
                    aws_secret_access_key   = config[f'aws_{env}']['aws_secret_access_key']
                )
        return cls.clients[env]


from ._imports import *
from pathlib import Path

def rmtree(root):
    '''remove folder and subfolder file and directory'''
    for p in root.iterdir():
        if p.is_dir():
            rmtree(p)
        else:
            p.unlink()

    root.rmdir()



__all__ = [
    's3_upload_file', 's3_save_and_upload_file', 's3_save_and_upload_csv', 's3_save_and_upload_json',
    's3_download_file', 's3_download_and_read_file', 's3_download_and_read_csv', 's3_download_and_read_json',
    's3_list_folder_contents', 's3_check_file_exists', 's3_download_folder_contents',
    's3_delete_file', 's3_sql_query_csv', 's3_upload_file_to_silver', 's3_get_folder_list', 's3_create_folder_list', 
    's3_compare_file_dates_in_silver', 's3_get_prediction_file_in_silver', 's3_remove_temp_local_download']

from ._imports import *
from ._config import config
from ._awsbase import Aws_base
from ._datetime import DtFind

class AmazonS3(Aws_base):
    clients = {}

    @classmethod
    def get_client(cls, env):
        return super()._get_client(env, 's3')

def client(env):
    return AmazonS3.get_client(env)

@atexit.register
def s3_close():
    for client in AmazonS3.clients.values():
        client.close()

def s3_upload_file(local_file_path, s3_file_path=None, overwrite=False, env='dev'):
    if s3_file_path == None:
        s3_file_path  = os.path.basename(local_file_path)
    if not overwrite and s3_file_path in s3_list_folder_contents(os.path.dirname(s3_file_path), env=env):
        raise Exception('File Already Exists')
    else:
        client(env).upload_file(local_file_path, config[f'aws_{env}']['bucket'], s3_file_path)

def s3_save_and_upload_file(data, s3_file_path, local_file_name='output', overwrite=False, save_file_kwargs={}, env='dev'):
    _, s3_file_extension = os.path.splitext(s3_file_path)
    local_folder_path    = config['local']['local_folder']
    local_file_name      = os.path.basename(local_file_name) + '.' + s3_file_extension
    local_file_path      = os.path.join(local_folder_path, local_file_name)
    if s3_file_extension == '.csv':
        data.to_csv(local_file_path, index=False, **save_file_kwargs)
    elif s3_file_extension == '.parquet':
        data.to_parquet(local_file_path, index=False, **save_file_kwargs)
    elif s3_file_extension == '.json':
        with open(local_file_path, 'w') as f:
            json.dump(data, f, indent=4, **save_file_kwargs)
    elif s3_file_extension == '.pkl':
        with open(local_file_path, 'wb') as f:
            pickle.dump(data, f)
    s3_upload_file(local_file_path, s3_file_path, overwrite=overwrite, env=env)
    os.remove(local_file_path)

def s3_save_and_upload_csv(df, local_file_name='output.csv', s3_file_path=None, overwrite=False, env='dev'):
    local_folder_path = config['local']['local_folder']
    local_file_name = os.path.basename(local_file_name)
    local_file_path = os.path.join(local_folder_path, local_file_name)
    df.to_csv(local_file_path, index=False)
    s3_upload_file(local_file_path, s3_file_path, overwrite=overwrite, env=env)
    os.remove(local_file_path)

def s3_save_and_upload_json(data, local_file_name='output.json', s3_file_path=None, overwrite=False, env='dev'):
    local_folder_path = config['local']['local_folder']
    local_file_name = os.path.basename(local_file_name)
    local_file_path = os.path.join(local_folder_path, local_file_name)
    with open(local_file_path, 'w') as f:
        json.dump(data, f, indent=4)
    s3_upload_file(local_file_path, s3_file_path, overwrite=overwrite, env=env)
    os.remove(local_file_path)
    
def s3_download_file(s3_file_path, local_file_path=None, s3_bucket=None, env='dev'):
    if s3_bucket is None:
        s3_bucket = config[f'aws_{env}']['bucket']
    if local_file_path is None:
        local_folder_path = config['local']['local_folder']
        local_file_name = os.path.basename(s3_file_path)
        local_file_path = os.path.join(local_folder_path, local_file_name)
    try:
        client(env).download_file(s3_bucket, s3_file_path, local_file_path)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == '404':
            raise Exception('File Does Not Exist')
        else:
            raise

def s3_download_and_read_file(s3_file_path, local_file_path=None, s3_bucket=None, read_file_kwargs={}, env='dev'):
    _, s3_file_extension = os.path.splitext(s3_file_path)
    if local_file_path is None:
        local_folder_path = config['local']['local_folder']
        local_file_name = os.path.basename(s3_file_path)
        local_file_path = os.path.join(local_folder_path, local_file_name)
    s3_download_file(s3_file_path, local_file_path, s3_bucket=s3_bucket, env=env)
    if s3_file_extension == '.csv':
        d = pd.read_csv(local_file_path, **read_file_kwargs)
    elif s3_file_extension == '.xlsx':
        d = pd.read_excel(local_file_path, **read_file_kwargs)
    elif s3_file_extension == '.parquet':
        d = pd.read_parquet(local_file_path, **read_file_kwargs)
    elif s3_file_extension == '.json':
        with open(local_file_path) as f:
            d = json.load(f)
    #os.remove(local_file_path)
    return d

def s3_download_and_read_csv(s3_file_path, local_file_path=None, s3_bucket=None, read_csv_kwargs={}, env='dev'):
    if local_file_path is None:
        local_folder_path = config['local']['local_folder']
        local_file_name = os.path.basename(s3_file_path)
        local_file_path = os.path.join(local_folder_path, local_file_name)
    s3_download_file(s3_file_path, local_file_path, s3_bucket=s3_bucket, env=env)
    df = pd.read_csv(local_file_path, **read_csv_kwargs)
    os.remove(local_file_path)
    return df

def s3_download_and_read_json(s3_file_path, local_file_path=None, s3_bucket=None, env='dev'):
    if local_file_path is None:
        local_folder_path = config['local']['local_folder']
        local_file_name = os.path.basename(s3_file_path)
        local_file_path = os.path.join(local_folder_path, local_file_name)
    s3_download_file(s3_file_path, local_file_path, s3_bucket=s3_bucket, env=env)
    with open(local_file_path) as f:
        d = json.load(f)
    os.remove(local_file_path)
    return d

def list_objects_with_paginator (bucket, prefix, env):
    '''
    list all file/folder within a prefix, can handle over 1000 files with paginator
    This is fix original function which is not able to handle over 1000 files in folder
    Parameters
    ----------
    Bucket : str 
                S3 bucket name.
    Prefix : str 
                S3 prefix.
    env : str 
                the env you are running.
    Returns
    ----------
    None : list 
                erturn list of folders/files path.
    Examples
    ----------
    list_objects_with_paginator (bucket_name, prefix, 'dev')
    '''
    s3_client = client(env)
    paginator = s3_client.get_paginator('list_objects')
    contents = []
    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):        
        if 'Contents' in page:
            for obj in page['Contents']:
                contents.append (obj['Key'])
    return contents

def s3_list_folder_contents(s3_folder_path, s3_bucket=None, pattern=r'^.*$', env='dev'):
    '''
    this funcition able to filter file/folder by regex expression 
    Parameters
    ----------
    s3_folder_path : str 
                s3 prefix.
    s3_bucket : str 
                s3 bucket name.
    pattern : str 
                regex pattern you want to match with file/folder name.
    env : str 
                environment you are running.
    Returns
    ----------
    contents : list 
                list of s3 path meeting the condition.
    Examples
    ----------
    s3_list_folder_contents(r'cash_flow_forecasting/model_assets', s3_bucket=None, pattern=f'^.*cash_flow_model_{'vendor'}_20250407_090547.csv$', env='dev')
    '''
    if s3_bucket is None:
        s3_bucket = config[f'aws_{env}']['bucket']
    contents = []
    for item in list_objects_with_paginator(bucket=s3_bucket, prefix=s3_folder_path, env =env):
        if re.match(pattern, item):
            contents.append(item)
    return contents


# def s3_list_folder_contents(s3_folder_path, s3_bucket=None, pattern=r'^.*$', env='dev'):
#     if s3_bucket is None:
#         s3_bucket = config[f'aws_{env}']['bucket']
#     contents = []
#     for item in client(env).list_objects(Bucket=s3_bucket, Prefix=s3_folder_path)['Contents']:
#         if re.match(pattern, item['Key']):
#             contents.append(item['Key'])
#     return contents

def s3_check_file_exists(s3_file_path, env='dev'):
    s3_folder_path = os.path.dirname(s3_file_path)
    if s3_file_path in s3_list_folder_contents(s3_folder_path, env=env):
        return True
    return False

def s3_download_folder_contents(s3_folder_path, local_folder_path=None, s3_bucket=None, pattern=r'^.*\.csv$', env='dev'):
    if local_folder_path is None:
        local_folder_path = config['local']['local_folder']
    for s3_file_path in s3_list_folder_contents(s3_folder_path, s3_bucket=s3_bucket, pattern=pattern, env=env):
        s3_file_name = os.path.basename(s3_file_path)
        local_file_path = os.path.join(local_folder_path, s3_file_name)
        s3_download_file(s3_file_path, local_file_path, s3_bucket=s3_bucket, env=env)

def s3_delete_file(s3_file_path, s3_bucket=None, env='dev'):
    if s3_bucket is None:
        s3_bucket = config[f'aws_{env}']['bucket']
    try:
        client(env).delete_object(Bucket=s3_bucket, Key=s3_file_path)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == '404':
            raise Exception('File Does Not Exist')
        else:
            raise

def s3_sql_query_csv(s3_file_path, s3_query='select * from s3object', s3_bucket=None, env='dev'):
    if s3_bucket is None:
        s3_bucket = config[f'aws_{env}']['bucket']
    resp = client(env).select_object_content(
        Bucket = s3_bucket,
        Key = s3_file_path,
        Expression=s3_query,
        ExpressionType='SQL',    
        InputSerialization={
            "CSV": {
                'FileHeaderInfo': 'USE',
                'AllowQuotedRecordDelimiter': True,
            },
        },
        OutputSerialization={
            'CSV': {}
        },
    )

    records = []
    for event in resp['Payload']:
        if 'Records' in event:
            records.append(event['Records']['Payload'])
        elif 'Stats' in event:
            stats = event['Stats']['Details']

    file_str = ''.join(r.decode('utf-8') for r in records)
    df = pd.read_csv(io.StringIO(file_str),  header=None)
    return df

def s3_create_single_folder(env= 'dev', s3_bucket = None, prefix = r'code_test/test/'):
    '''
    this is use to programmatically generate folder
    s3_create_single_folder(env= 'dev', s3_bucket = None, prefix = r'code_test/test/')
    '''
    if not s3_bucket:
        s3_bucket = config[f'aws_{env}']['bucket']
    client(env).put_object(Bucket=s3_bucket, Key=(prefix))
    return 

def s3_convert_file_to_df ( bucket, file_path, env):
    
    s3_client = boto3.client('s3',
                            region_name='us-east-2',aws_access_key_id=config[f'aws_{env}']['aws_access_key_id'],aws_secret_access_key=config[f'aws_{env}']['aws_secret_access_key']
                            )
    obj = s3_client.get_object(Bucket=bucket, Key=file_path)
    output = pd.read_csv(io.BytesIO(obj['Body'].read()),index_col=False)
    return output

def s3_upload_file_to_silver(local_file_path, s3_file_path=None, overwrite=False, env='dev'):
    
    if s3_file_path == None:
        s3_file_path  = os.path.basename(local_file_path)
    if not overwrite and s3_file_path in s3_list_folder_contents(os.path.dirname(s3_file_path), env=env):
        raise Exception('File Already Exists')
    else:
        client(env).upload_file(local_file_path,f'amica-aws-eds-{env}-dl-silver' , s3_file_path)


#copy folder structure
def s3_get_folder_list( env = 'dl', s3_folder_path = "archive/", s3_bucket = None,  pattern = r'^.*$'):
    if not s3_bucket:        
        s3_bucket = config[f'aws_{env}']['bucket']

    contents = set()    
    for item in client(env).list_objects(Bucket=s3_bucket, Prefix=s3_folder_path)['Contents']:
        contents.add(os.path.dirname(item['Key']))  #this only get folder name
    return contents

def s3_create_folder_list (folder_list, env = 'dl', original_project_prefix = 'archive', new_project_prefix = "archive1", s3_bucket = 'amica-eds-edsc'):
    #https://stackoverflow.com/questions/1939743/amazon-s3-boto-how-to-create-a-folder
    # s3 = boto3.resource('s3')
    # bucket = s3.Bucket('amica-eds-edsc')
    #project_path = r"archive1/"
    '''
    archive_list2 = s3_get_folder_list( env = 'dl', s3_folder_path = r"archive/drtest/", s3_bucket = None,  pattern = r'^.*$')
    
    '''
    for k in folder_list:
        #remove the old name
        k = k.replace(original_project_prefix, new_project_prefix, 1)  #only replace the first occurance
        prefix = f'{k}/'        
        client(env).put_object(Bucket=s3_bucket, Key=(prefix))
    return 

def s3_delete_all_objects_from_folder(bucket_name = "amica-eds-edsc", prefix = r"archive/drtest2/", env = 'dl'):
    """
    This function deletes all files in a folder from S3 bucket for troubleshooting purpose
    s3_delete_all_objects_from_folder(bucket_name = "amica-eds-edsc", prefix = r"archive/drtest2/", env = 'dl')
    """
    
    s3_client = client(env)
    # First we list all files in folder
    response = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
    files_in_folder = response["Contents"]
    files_to_delete = []
    # We will create Key array to pass to delete_objects function
    for f in files_in_folder:
        files_to_delete.append({"Key": f["Key"]})
    # This will delete all files in a folder
    s3_client.delete_objects(
        Bucket=bucket_name, Delete={"Objects": files_to_delete}
    )
    return 

#this is for checking if we need to rerun on business day 
def s3_compare_file_dates_in_silver(env='dev'):
    '''
    this function is used to avoid repeated run in prod because accounting required using business first day, 
    we need to schedule run prediction multi time to accomodate holidy and weekend. To check if prediction has been run  
    check prediction file in silver latest time vs expected time to determine next step
    actual_file_year_month, expected_file_year_month = compare_file_dates_in_silver(env='dev')
    print(actual_file_year_month, expected_file_year_month)
    '''
    today = dt.datetime.now()
    objects = client(env).list_objects(Bucket=f'amica-aws-eds-{env}-dl-silver', Prefix = r'cashflow_forecast/forecast_stage/') 
    df = pd.DataFrame(objects["Contents"])
    actual_file_year_month =  (DtFind.first_day_last_month(today).strftime('%Y-%m') if not df[~df['Key'].str.endswith(r'/')]['LastModified'].to_list()  #if no file, get last month 
                            else df['LastModified'].dt.tz_convert('US/Eastern').apply(pd.to_datetime).max().strftime("%Y-%m") #if contain file, get max date )

    )
    expected_file_year_month = today.strftime("%Y-%m")
    return actual_file_year_month, expected_file_year_month

def s3_get_prediction_file_in_silver (s3_folder_path = r"cashflow_forecast/forecast_stage/", s3_bucket = f'amica-aws-eds-dev-dl-silver', pattern = r'^.*\.parquet$', env = 'dev'):
    file_list = []
    for item in client(env).list_objects(Bucket=s3_bucket, Prefix=s3_folder_path)['Contents']:
        if item['Key'][-1] != "/"  and  re.match(pattern, item['Key']):     
            file_list.append(item['Key'])
    return file_list

def s3_remove_temp_local_download (s3_file_path):
    '''
    this help remove local file downloaded from s3, based on s3 path    
    '''
    local_folder_path = config['local']['local_folder']
    local_file_path = os.path.join(local_folder_path, os.path.basename(s3_file_path))
    os.remove(local_file_path)
    return


from cffds.projects.life_cash_flow_forecasting import *
from tabulate import tabulate
logfile, logger = create_logger()

env = sys.argv[1]
#test current code end to end, see if there is error 
cf_types = ['vendor','salary', 'claim loss', 'claim subrogation', 'inflow premiums','inflow receipts','outflow premiums',]
#cf_types = ['vendor','salary',  'inflow premiums','outflow premiums',]
df_job_status = pd.DataFrame (columns = ['Job Type', 'Status',  'Error Info' ])

if env =='dev':
    train_ends = ['2025-02-01']*len(cf_types)   #because dev env, data is not complete, for test purpose
    #train_ends = len(cf_types)*[None]  #we will run scheduled job in dev for few time like prod 
elif env =='prod':
    train_ends = len(cf_types)*[None]

#this is used to determine if job need to run as we schedule 4 time to only run on business day
run_flag = 0
actual_file_year_month, expected_file_year_month = s3_compare_file_dates_in_silver(env)
if env == 'prod':   
    if (expected_file_year_month  >  actual_file_year_month):
        run_flag = 1
    else:
        logger.info(f'''Please check {env} environment silver prediction folder 
                        expect run year month is {expected_file_year_month},
                        maxium year month aleady generated file is {actual_file_year_month}
                    ''')
else:  #dev will run without this limit
    run_flag = 1

if run_flag == 1:
    for t, te in zip(cf_types, train_ends):
        try:
            error_status = 0
            error_status = scoring_pipeline(cf_type = t, env = env, logger=logger, train_end=te)
            #save success info
            logger.info(f'job {t} completed successfully!')
            df_job_status.loc[len(df_job_status)] = [t, 'Success' if error_status ==0 else 'Fail', None]
        except Exception as e:
            error_status = 1
            logger.info(f'job {t} has error')
            logger.error('Error', exc_info=e)
            df_job_status.loc[len(df_job_status)] = [t, 'Fail', e]
    
    save_latest_prediction (env)
    save_monitor_dashboard_value (env)
    sns_sub =  f"Data science team cash flow job status in {env}"
    sns_msg = tabulate(df_job_status, headers='keys', tablefmt='psql')
    if env =='prod':
        send_sns(sns_sub, sns_msg, env)
    else:
        print(sns_msg)
    logger.info(sns_msg)
save_logger(logfile, env=env)


__all__ = ['admission', 'score_pipeline']

from ._imports import *
from ._pconfig import *
from ._features import *
from ._monitor import *
from ._train import train_pipeline, dataset_initialize, dataset_filter, load_model_config, load_model
from ._misc import *

def admission():
    print('testing...')

def score_pipeline(run_month, run_day=pconfig['run_day'], env='dev', logger=None):
    error_status = 0
    if logger:
        logger.info(f'Score for {run_month} performed on date {run_day}.')
    try:
        runtime = create_runtime(run_month, run_day, env, logger)
        df_base = dataset_initialize(runtime)
        df_feat = dataset_features(df_base, runtime)
        df_filt = dataset_filter(df_feat, runtime)
        df_pred = dataset_score(df_filt, runtime)

        save_output_eds(df_pred, runtime)
        save_output_care(df_pred, runtime)

        model_config = load_model_config(runtime)
        trackers     = monitor_pipeline(df_pred, model_config, logger=logger, env=env)
        trackers['error'] = error_status
        
        if trackers['alert'] or run_month.endswith('-01'):
            # if check_previous_retrain(runtime):
            #     update_model_version(env=env, promote=False)
            # else:
            today_dt     = dt.datetime.today()
            train_start  = DtFind.first_day_of_month(today_dt - dt.timedelta(days=8*30), '%Y-%m')
            train_end    = DtFind.first_day_of_month(today_dt - dt.timedelta(days=5*30), '%Y-%m')
            if logger:
                logger.info(f'Train from {train_start} to {train_end}.')
            train_pipeline(train_start, train_end, env=env, logger=logger, update_model=True)


    except Exception as e:
        if logger:
            logger.error('Error', exc_info=e)
        error_status = 1
        trackers     = {'error': error_status}
    finally:
        save_output_mstr(trackers, runtime)
        return error_status

def create_runtime(run_month, run_day, env, logger):
    timestamp = dt.datetime.now().strftime("%Y%m%d%H%M%S")
    if logger:
        logger.info(f'Run timestamp: {timestamp}')

    features = Features()
    features.remove_features(['CANCEL_RENEWAL', 'CANCEL_EXPIRING'])
    features.remove_features(['CC_CLAIM_NOTES'])
    #remove demographic 
    demo_feature = ['GOOD_STUDENT','NUM_CONVICTIONS','DRIVER_AGE_DISC','DRIVER_TRAINING','YOUTHFUL_OPER','STUDENT_AWAY','MILITARY_IND','RIDER_TRAINING','HOUSEHOLD_INCOME']
    features.remove_features(demo_feature)
    # features.remove_features([c for c in features.get_feature_dict() if c.startswith('PC') or c.startswith('CC')]) # debugging purpose for now

    runtime = {}
    runtime['timestamp']      = timestamp
    runtime['env']            = env
    runtime['features']       = features.get_feature_dict()
    runtime['teradata_table'] = f'{config["teradata"]["user"]}.temp_table'
    runtime['type_dt']        = pconfig['type_dt']

    run_month = dt.datetime.strptime(run_month, '%Y-%m')
    runtime['run_month']       = run_month
    runtime['run_month_start'] = DtFind.first_day_of_month(run_month).strftime('%Y-%m-%d')
    runtime['run_month_end']   = DtFind.last_day_of_month(run_month).strftime('%Y-%m-%d')
    if str(run_day).isnumeric():
        runtime['run_dt']      = DtFind.first_day_last_month(run_month).replace(day=run_day).strftime('%Y-%m-%d')
    else: # allow passing in '2023-04-27' as run day for example
        runtime['run_dt']      = dt.datetime.strptime(run_day,'%Y-%m-%d').strftime('%Y-%m-%d')
    return runtime

def dataset_score(df, runtime):
    model_config = load_model_config(runtime)
    model        = load_model(model_config, runtime)

    df_feat  = df[model_config['features']]
    probs    = model.predict_proba(df_feat)[:,1]
    preds    = (probs>float(model_config['threshold'])).astype(int)
    df_score = df.assign(
        MODEL_VERSION = model_config['model_version'],
        THRESHOLD     = model_config['threshold'],
        PROBABILITY   = probs,
        PREDICTION    = preds,
    )
    return df_score

def check_previous_retrain(runtime, window=30, dt_format='%Y-%m-%d'):
    fnames = s3_list_folder_contents(os.path.join(
        pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
        'retrain/'
    ), pattern=r'^.*cancel_model_update.*\.json$')
    fnames = pd.DataFrame({'file': fnames})
    if len(fnames)==0:
        return False
    pat    = r'cancel_model_update_([\d]*)\.json'
    fnames['timestamp'] = fnames['file'].str.extract(pat)
    fnames['run_date']  = pd.to_datetime(fnames['timestamp']).dt.strftime('%Y-%m-%d')
    
    for _, row in fnames.iterrows():
        trackers = s3_download_and_read_json(row['file'])
        trackers_df = pd.Series(trackers)
        fnames.loc[fnames['file']==row['file'], trackers_df.index] = trackers_df.values
    fnames = fnames.loc[fnames['action']=='promote']
    fnames = fnames.sort_values(by=['datetime'])
    if len(fnames)==0:
        return False
    promotion_date = fnames.iloc[-1].to_dict()['datetime']

    run_date       = dt.datetime.strptime(runtime['run_dt'], dt_format)
    promotion_date = dt.datetime.strptime(promotion_date, dt_format)
    return promotion_date > run_date - dt.timedelta(window)

def save_output_eds(df, runtime):
    s3_save_and_upload_csv(
        df, 
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_eds_folder_{runtime["env"]}'],
            f'cancel_{runtime["run_month_start"]}_{runtime["run_dt"]}_{runtime["timestamp"]}.csv'
        ),
        env=runtime['env'],
    )

def save_output_care(df, runtime):
    df = df[['POLICY_NO', 'ACCOUNT_NO', 'PREDICTION', 'PROBABILITY', 'THRESHOLD', 'EFFECTIVE_DATE', 'EXPIRATION_DATE']]
    df['ID']             = 'Retention_' + df['POLICY_NO']
    df['Source']         = 'Retention'
    df['RecCreatedDate'] = dt.datetime.strptime(runtime['timestamp'], "%Y%m%d%H%M%S")
    df['PREDICTION']     = df['PREDICTION'].apply(lambda r: 'ActionableCancel' if r>0 else 'NotActionableCancel')
    df = df.rename(columns={
        'POLICY_NO': 'PolicyNumber', 'ACCOUNT_NO': 'AccountNumber',
        'EFFECTIVE_DATE': 'PolicyEffectiveDate', 'EXPIRATION_DATE': 'PolicyExpiredDate',
        'PREDICTION': 'Prediction', 'PROBABILITY': 'Probability', 'THRESHOLD': 'Threshold',
    })
    df = df[['ID', 'Source', 'PolicyNumber', 'AccountNumber', 'Prediction', 'Probability', 'Threshold', 'PolicyEffectiveDate', 'PolicyExpiredDate', 'RecCreatedDate']]

    s3_save_and_upload_csv(
        df, 
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_care_folder_{runtime["env"]}'],
            f'cancel_{runtime["run_month_start"]}_{runtime["timestamp"]}.csv'
        ),
        env=runtime['env'],
    )

    s3_save_and_upload_csv(
        df, 
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_care_folder_{runtime["env"]}'],
            'prediction_current.csv'
        ),
        overwrite=True,
        env=runtime['env'],
    )

    if any([(os.path.basename(f)=='prediction_history.csv') for f in s3_list_folder_contents(pconfig[f's3_model_output_care_folder_{runtime["env"]}'])]):
        df_history = s3_download_and_read_csv(os.path.join(
            pconfig[f's3_model_output_care_folder_{runtime["env"]}'],
            'prediction_history.csv'
        ), env=runtime['env'])
        df_history['RecCreatedDate'] = pd.to_datetime(df_history['RecCreatedDate'], errors='coerce')
        df_history = pd.concat([df_history, df], ignore_index=True)
        df_history = (
            df_history.sort_values(by='RecCreatedDate', ascending=False)
            .drop_duplicates(subset=['PolicyNumber', 'AccountNumber', 'PolicyEffectiveDate'], keep='first')
        )
        s3_save_and_upload_csv(
            df_history, 
            s3_file_path = os.path.join(
                pconfig[f's3_model_output_care_folder_{runtime["env"]}'],
                'prediction_history.csv'
            ),
            overwrite=True,
            env=runtime['env'],
        )
        
    else:
        s3_save_and_upload_csv(
            df, 
            s3_file_path = os.path.join(
                pconfig[f's3_model_output_care_folder_{runtime["env"]}'],
                'prediction_history.csv'
            ),
            overwrite=False,
            env=runtime['env'],
        )

def save_output_mstr(data, runtime):
    s3_save_and_upload_json(
        data,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
            'monitor/',
            f'cancel_{runtime["run_dt"]}_{runtime["timestamp"]}.json'
        ),
        env=runtime['env'],
    )